#!/usr/bin/python3
#
# Copyright 2013-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0

import argparse
import os
import pdb
import re
import sys
import time
import xml.etree.ElementTree as etree

sys.path.append(os.path.abspath(os.path.dirname(__file__)))

from cgenerator import CGeneratorOptions, COutputGenerator
from creflectiongenerator import CReflectionOutputGenerator
from docgenerator import DocGeneratorOptions, DocOutputGenerator
from interfacedocgenerator import InterfaceDocGenerator
from extensionmetadocgenerator import (ExtensionMetaDocGeneratorOptions,
                                       ExtensionMetaDocOutputGenerator)
from generator import write
from hostsyncgenerator import HostSynchronizationOutputGenerator
from indexgenerator import DocIndexOutputGenerator
from pygenerator import PyOutputGenerator
from rubygenerator import RubyOutputGenerator
from reflib import logDiag, logWarn, logErr, setLogFile
from reg import Registry
from validitygenerator import ValidityOutputGenerator
from apiconventions import APIConventions

# Simple timer functions
startTime = None


def startTimer(timeit):
    global startTime
    if timeit:
        startTime = time.process_time()


def endTimer(timeit, msg):
    global startTime
    if timeit and startTime is not None:
        endTime = time.process_time()
        logDiag(msg, endTime - startTime)
        startTime = None


def makeREstring(strings, default=None, strings_are_regex=False):
    """Turn a list of strings into a regexp string matching exactly those strings."""
    if strings or default is None:
        if not strings_are_regex:
            strings = (re.escape(s) for s in strings)
        return f"^({'|'.join(strings)})$"
    return default


def makeGenOpts(args):
    """Returns a directory of [ generator function, generator options ] indexed
    by specified short names. The generator options incorporate the following
    parameters:

    args is an parsed argument object; see below for the fields that are used."""
    global genOpts
    genOpts = {}

    # Default class of extensions to include, or None
    defaultExtensions = args.defaultExtensions

    # Additional extensions to include (list of extensions)
    extensions = args.extension

    # Extensions to remove (list of extensions)
    removeExtensions = args.removeExtensions

    # Extensions to emit (list of extensions)
    emitExtensions = args.emitExtensions

    # Extension to emit for standalone_header (single extension)
    standaloneExtension = args.standaloneExtension

    # Header prefix override for standalone headers (list of strings)
    standalonePrefixOverride = args.standalonePrefixOverride

    # Features to include (list of features)
    features = args.feature

    # Whether to disable inclusion protect in headers
    protect = args.protect

    # Output target directory
    directory = args.directory

    # Path to generated files, particularly apimap.py
    genpath = args.genpath

    # Descriptive names for various regexp patterns used to select
    # versions and extensions
    allFeatures = allExtensions = r'.*'

    # Turn lists of names/patterns into matching regular expressions
    addExtensionsPat     = makeREstring(extensions, None)
    removeExtensionsPat  = makeREstring(removeExtensions, None)
    emitExtensionsPat    = makeREstring(emitExtensions, allExtensions)
    featuresPat          = makeREstring(features, allFeatures)

    # REUSE-IgnoreStart
    # Copyright text prefixing all headers (list of strings).
    prefixStrings = [
        '/*',
        '** Copyright 2017-2024, The Khronos Group Inc.',
        '**',
        # The following split string is to avoid confusing the "REUSE" tool
        '** SPDX-License-Identifier' + ': Apache-2.0 OR MIT',
        '*/',
        ''
    ]
    # REUSE-IgnoreEnd

    # Text specific to OpenXR headers
    xrPrefixStrings = [
        '/*',
        '** This header is generated from the Khronos OpenXR XML API Registry.',
        '**',
        '*/',
        ''
    ]

    # Include the non-platform openxr header in the platform header.
    platformPrefixStrings = [
        '#include "openxr.h"'
    ]

    # Defaults for generating re-inclusion protection wrappers (or not)
    protectFile = protect

    # An API style conventions object
    conventions = APIConventions()

    # OpenXR 1.0 - header for core API + extensions.
    # To generate just the core API,
    # change to 'defaultExtensions = None' below.
    genOpts['openxr.h'] = [
          COutputGenerator,
          CGeneratorOptions(
            conventions       = conventions,
            filename          = 'openxr.h',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = 'openxr',
            addExtensions     = None,
            removeExtensions  = None,
            emitExtensions    = emitExtensionsPat,
            prefixText        = prefixStrings + xrPrefixStrings,
            genFuncPointers   = True,
            protectFile       = protectFile,
            protectFeature    = False,
            protectProto      = '#ifndef',
            protectProtoStr   = 'XR_NO_PROTOTYPES',
            protectExtensionProto      = '#ifdef',
            protectExtensionProtoStr   = 'XR_EXTENSION_PROTOTYPES',
            apicall           = 'XRAPI_ATTR ',
            apientry          = 'XRAPI_CALL ',
            apientryp         = 'XRAPI_PTR *',
            alignFuncParam    = 48,
            genAliasMacro     = True,
            genStructExtendsComment = True,
            aliasMacro        = 'XR_MAY_ALIAS')
        ]

    standalonePrefixString = prefixStrings + xrPrefixStrings
    if len(standalonePrefixOverride) > 0:
        standalonePrefixString = standalonePrefixOverride

    # "XR_EXT_some_extension" becomes "ext_some_extension.h"
    standaloneExtensionFileName = f"{standaloneExtension[len('XR_'):].lower()}.h"

    genOpts['standalone_header'] = [
          COutputGenerator,
          CGeneratorOptions(
            conventions       = conventions,
            filename          = standaloneExtensionFileName,
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = None,
            emitversions      = None,
            emitComments      = True,
            defaultExtensions = 'openxr',
            addExtensions     = None,
            removeExtensions  = None,
            emitExtensions    = standaloneExtension,
            prefixText        = standalonePrefixString,
            genFuncPointers   = True,
            protectFile       = protectFile,
            protectFeature    = True,
            protectProto      = '#ifndef',
            protectProtoStr   = 'XR_NO_PROTOTYPES',
            protectExtensionProto      = '#ifdef',
            protectExtensionProtoStr   = 'XR_EXTENSION_PROTOTYPES',
            apicall           = 'XRAPI_ATTR ',
            apientry          = 'XRAPI_CALL ',
            apientryp         = 'XRAPI_PTR *',
            alignFuncParam    = 48,
            genAliasMacro     = True,
            genStructExtendsComment = True,
            aliasMacro        = 'XR_MAY_ALIAS',
            misracppstyle     = False,
            reparentEnums     = False,
            # Crucial part of making standalone headers is adding struct types
            # that would otherwise be added to openxr.h
            redefineEnumExtends = True,
            # Ensure we only include the directly
            # required types, and not recursively
            # required ones. Otherwise it would
            # create collisions with types defined
            # in main openxr.h
            emitRecursiveRequirements = False,
            )
        ]


    # OpenXR platform header for Graphics API and Platform extensions.
    genOpts['openxr_platform.h'] = [
          COutputGenerator,
          CGeneratorOptions(
            conventions       = conventions,
            filename          = 'openxr_platform.h',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = 'openxr',
            addExtensions     = None,
            removeExtensions  = None,
            emitExtensions    = emitExtensionsPat,
            prefixText        = prefixStrings + xrPrefixStrings + platformPrefixStrings,
            genFuncPointers   = True,
            protectFile       = protectFile,
            protectFeature    = False,
            protectProto      = '#ifndef',
            protectProtoStr   = 'XR_NO_PROTOTYPES',
            protectExtensionProto      = '#ifdef',
            protectExtensionProtoStr   = 'XR_EXTENSION_PROTOTYPES',
            apicall           = 'XRAPI_ATTR ',
            apientry          = 'XRAPI_CALL ',
            apientryp         = 'XRAPI_PTR *',
            alignFuncParam    = 48,
            genAliasMacro     = True,
            genStructExtendsComment = True,
            aliasMacro        = 'XR_MAY_ALIAS')
    ]

    genOpts['openxr_loader_negotiation.h'] = [
          COutputGenerator,
          CGeneratorOptions(
            conventions       = conventions,
            filename          = 'openxr_loader_negotiation.h',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = 'openxr',
            addExtensions     = None,
            removeExtensions  = None,
            emitExtensions    = emitExtensionsPat,
            prefixText        = prefixStrings + xrPrefixStrings + platformPrefixStrings,
            genFuncPointers   = True,
            protectFile       = protectFile,
            protectFeature    = False,
            protectProto      = '#ifndef',
            protectProtoStr   = 'XR_NO_PROTOTYPES',
            protectExtensionProto      = '#ifdef',
            protectExtensionProtoStr   = 'XR_EXTENSION_PROTOTYPES',
            apicall           = 'XRAPI_ATTR ',
            apientry          = 'XRAPI_CALL ',
            apientryp         = 'XRAPI_PTR *',
            alignFuncParam    = 48,
            genAliasMacro     = True,
            genStructExtendsComment = True,
            aliasMacro        = 'XR_MAY_ALIAS')
    ]

    def make_reflection_options(fn):
        return CGeneratorOptions(
            conventions=conventions,
            filename=fn,
            directory=directory,
            apiname='openxr',
            profile=None,
            versions=featuresPat,
            emitversions=featuresPat,
            defaultExtensions='openxr',
            addExtensions=None,
            removeExtensions=None,
            emitExtensions=emitExtensionsPat,
            prefixText=prefixStrings + xrPrefixStrings + platformPrefixStrings,
            genFuncPointers=True,
            protectFile=protectFile,
            protectFeature=False,
            protectProto='#ifndef',
            protectProtoStr='XR_NO_PROTOTYPES',
            apicall='XRAPI_ATTR ',
            apientry='XRAPI_CALL ',
            apientryp='XRAPI_PTR *',
            alignFuncParam=48,
            genAliasMacro=True,
            genStructExtendsComment=True,
            aliasMacro='XR_MAY_ALIAS')

    # OpenXR generic reflection header
    genOpts['openxr_reflection.h'] = [
        CReflectionOutputGenerator,
        make_reflection_options('openxr_reflection.h'),
    ]

    genOpts['openxr_reflection_structs.h'] = [
        CReflectionOutputGenerator,
        make_reflection_options('openxr_reflection_structs.h'),
    ]

    genOpts['openxr_reflection_parent_structs.h'] = [
        CReflectionOutputGenerator,
        make_reflection_options('openxr_reflection_parent_structs.h'),
    ]

    # OpenXR 1.1 - API include files for spec and ref pages
    # Overwrites include subdirectories in spec source tree
    # The generated include files do not include the calling convention
    # macros (apientry etc.), unlike the header files.
    # Because the 1.1 main branch includes ref pages for extensions,
    # all the extension interfaces need to be generated, even though
    # none are used by the core spec itself.
    genOpts['apiinc'] = [
          DocOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'apiinc',
            directory         = directory,
            genpath           = genpath,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat,
            prefixText        = prefixStrings + xrPrefixStrings,
            apicall           = '',
            apientry          = '',
            apientryp         = '*',
            alignFuncParam    = 48,
            secondaryInclude  = True,
            expandEnumerants  = False,
            extEnumerantAdditions = True)
        ]

    # Python and Ruby representations of API information, used by scripts
    # that do not need to load the full XML.
    genOpts['apimap.py'] = [
          PyOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'apimap.py',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat)
        ]

    # Python and Ruby representations of API information, used by scripts
    # that do not need to load the full XML.
    genOpts['apimap.rb'] = [
          RubyOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'apimap.rb',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat)
        ]

    # Index chapter
    genOpts['index.adoc'] = [
          DocIndexOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'index.adoc',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat)
        ]

    # Core API validity files for spec
    genOpts['validinc'] = [
          ValidityOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'validinc',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat)
        ]

    # Core API host sync table files for spec
    genOpts['hostsyncinc'] = [
          HostSynchronizationOutputGenerator,
          DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'hostsyncinc',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat)
        ]

    # Extension metainformation for spec extension appendices
    genOpts['extinc'] = [
          ExtensionMetaDocOutputGenerator,
          ExtensionMetaDocGeneratorOptions(
            conventions       = conventions,
            filename          = 'extinc',
            directory         = directory,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = None,
            defaultExtensions = defaultExtensions,
            addExtensions     = None,
            removeExtensions  = None,
            emitExtensions    = emitExtensionsPat)
        ]
    # Version and extension interface docs for version/extension appendices
    # Includes all extensions by default.
    genOpts['interfaceinc'] = [
        InterfaceDocGenerator,
        DocGeneratorOptions(
            conventions       = conventions,
            filename          = 'interfaceinc',
            directory         = directory,
            genpath           = None,
            apiname           = 'openxr',
            profile           = None,
            versions          = featuresPat,
            emitversions      = featuresPat,
            defaultExtensions = None,
            addExtensions     = addExtensionsPat,
            removeExtensions  = removeExtensionsPat,
            emitExtensions    = emitExtensionsPat,
            reparentEnums     = False)
        ]

def genTarget(args):
    """Create an API generator and corresponding generator options based on
    the requested target and command line options.

    This is encapsulated in a function so it can be profiled and/or timed.
    The args parameter is an parsed argument object containing the following
    fields that are used:

    - target - target to generate
    - directory - directory to generate it in
    - protect - True if re-inclusion wrappers should be created
    - extensions - list of additional extensions to include in generated interfaces"""

    # Create generator options with specified parameters
    makeGenOpts(args)

    # Select a generator matching the requested target
    if args.target in genOpts:
        createGenerator = genOpts[args.target][0]
        options = genOpts[args.target][1]

        if not args.quiet:
            write('* Building', options.filename, file=sys.stderr)
            write('* options.versions          =', options.versions, file=sys.stderr)
            write('* options.emitversions      =', options.emitversions, file=sys.stderr)
            write('* options.defaultExtensions =', options.defaultExtensions, file=sys.stderr)
            write('* options.addExtensions     =', options.addExtensions, file=sys.stderr)
            write('* options.removeExtensions  =', options.removeExtensions, file=sys.stderr)
            write('* options.emitExtensions    =', options.emitExtensions, file=sys.stderr)



        startTimer(args.time)
        gen = createGenerator(errFile=errWarn,
                              warnFile=errWarn,
                              diagFile=diag)
        return (gen, options)
    else:
        write('No generator options for unknown target:',
              args.target, file=sys.stderr)
        return None


# -feature name
# -extension name
# For both, "name" may be a single name, or a space-separated list
# of names, or a regular expression.
if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    parser.add_argument('-defaultExtensions', action='store',
                        default='openxr',
                        help='Specify a single class of extensions to add to targets')
    parser.add_argument('-extension', action='append',
                        default=[],
                        help='Specify an extension or extensions to add to targets')
    parser.add_argument('-removeExtensions', action='append',
                        default=[],
                        help='Specify an extension or extensions to remove from targets')
    parser.add_argument('-emitExtensions', action='append',
                        default=[],
                        help='Specify an extension or extensions to emit in targets')
    parser.add_argument('-standaloneExtension', action='store',
                        default='',
                        help='Specify an extension to emit during standalone_header generation')
    parser.add_argument('-standalonePrefixOverride', action='append',
                        default=[],
                        help='Override the header prefix of headers generated with "standalone_header"')

    parser.add_argument('-feature', action='append',
                        default=[],
                        help='Specify a core API feature name or names to add to targets')
    parser.add_argument('-debug', action='store_true',
                        help='Enable debugging')
    parser.add_argument('-dump', action='store_true',
                        help='Enable dump to stderr')
    parser.add_argument('-diagfile', action='store',
                        default=None,
                        help='Write diagnostics to specified file')
    parser.add_argument('-errfile', action='store',
                        default=None,
                        help='Write errors and warnings to specified file instead of stderr')
    parser.add_argument('-noprotect', dest='protect', action='store_false',
                        help='Disable inclusion protection in output headers')
    parser.add_argument('-profile', action='store_true',
                        help='Enable profiling')
    parser.add_argument('-registry', action='store',
                        default='xr.xml',
                        help='Use specified registry file instead of xr.xml')
    parser.add_argument('-time', action='store_true',
                        help='Enable timing')
    parser.add_argument('-genpath', action='store', default='generated',
                        help='Path to generated files')
    parser.add_argument('-o', action='store', dest='directory',
                        default='.',
                        help='Create target and related files in specified directory')
    parser.add_argument('target', metavar='target', nargs='?',
                        help='Specify target')
    parser.add_argument('-quiet', action='store_true', default=False,
                        help='Suppress script output during normal execution.')
    parser.add_argument('-verbose', action='store_false', dest='quiet', default=True,
                        help='Enable script output during normal execution.')

    args = parser.parse_args()

    # This splits arguments which are space-separated lists
    args.feature = [name for arg in args.feature for name in arg.split()]
    args.extension = [name for arg in args.extension for name in arg.split()]

    # create error/warning & diagnostic files
    if args.errfile:
        errWarn = open(args.errfile, 'w', encoding='utf-8')
    else:
        errWarn = sys.stderr

    if args.diagfile:
        diag = open(args.diagfile, 'w', encoding='utf-8')
    else:
        diag = None

    if args.time:
        # Log diagnostics and warnings
        setLogFile(setDiag = True, setWarn = True, filename = '-')

    # Create the API generator & generator options
    (gen, options) = genTarget(args)

    # Create the registry object with the specified generator and generator
    # options. The options are set before XML loading as they may affect it.
    reg = Registry(gen, options)

    # Parse the specified registry XML into an ElementTree object
    startTimer(args.time)
    tree = etree.parse(args.registry)
    endTimer(args.time, '* Time to make ElementTree =')

    # Load the XML tree into the registry object
    startTimer(args.time)
    reg.loadElementTree(tree)
    endTimer(args.time, '* Time to parse ElementTree =')

    if args.dump:
        write('* Dumping registry to regdump.txt', file=sys.stderr)
        reg.dumpReg(filehandle=open('regdump.txt', 'w', encoding='utf-8'))

    # Finally, use the output generator to create the requested target
    if args.debug:
        pdb.run('reg.apiGen()')
    else:
        startTimer(args.time)
        reg.apiGen()
        endTimer(args.time, f"* Time to generate {options.filename} =")

    if not args.quiet:
        logDiag('* Generated', options.filename)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2020 Valve Corporation
# Copyright (c) 2018-2023 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import json
import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write
from khronos_replay_consumer_body_generator import KhronosReplayConsumerBodyGenerator


class VulkanReplayDumpResourcesBodyGeneratorOptions(VulkanBaseGeneratorOptions):
    """Options for generating a C++ class for Vulkan capture file replay."""

    def __init__(
        self,
        dump_resources_overrides=None,  # Path to JSON file listing Vulkan API calls to override on replay.
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )
        self.dump_resources_overrides = dump_resources_overrides

        self.begin_end_file_data.specific_headers.extend((
            'generated/generated_vulkan_replay_dump_resources.h',
            '',
            'util/defines.h',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))
        self.begin_end_file_data.common_api_headers = []


class VulkanReplayDumpResourcesBodyGenerator(
    KhronosReplayConsumerBodyGenerator, VulkanBaseGenerator
):
    """VulkanReplayDumpResourcesBodyGenerator - subclass of VulkanBaseGenerator.
    """

    # Map of Vulkan function names to override function names.  Calls to Vulkan functions in the map
    # will be replaced by the override value.
    DUMP_RESOURCES_OVERRIDES = {}

    # Map of pool object types associating the pool type with the allocated type and the allocated type with the pool type.
    POOL_OBJECT_ASSOCIATIONS = {
        'VkCommandBuffer': 'VkCommandPool',
        'VkDescriptorSet': 'VkDescriptorPool',
        'VkCommandPool': 'VkCommandBuffer',
        'VkDescriptorPool': 'VkDescriptorSet'
    }

    NOT_SKIP_FUNCTIONS_OFFSCREEN = ['Create', 'Destroy', 'GetSwapchainImages', 'AcquireNextImage', 'QueuePresent']

    SKIP_FUNCTIONS_OFFSCREEN = ['Surface', 'Swapchain', 'Present']

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def endFile(self):
        """Method override."""
        api_data = self.get_api_data()

        KhronosReplayConsumerBodyGenerator.generate_replay_consumer_content(self, api_data)

        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def make_consumer_func_body(self, api_data, return_type, name, values):
        """
        Method override.
        Return VulkanReplayConsumer class member function definition.
        """
        body = ''

        is_override = name in self.DUMP_RESOURCES_OVERRIDES

        if not is_override:
            body += '    if (IsRecording(commandBuffer))\n'
            body += '    {\n'
            body += '        CommandBufferIterator first, last;\n'
            body += '        bool found = GetDrawCallActiveCommandBuffers(commandBuffer, first, last);\n'
            body += '        if (found)\n'
            body += '        {\n'
            body += '            for (CommandBufferIterator it = first; it < last; ++it)\n'
            body += '            {\n'

            dispatchfunc = 'func(*it, '

            call_expr = ''
            for val in values[1:]:
                call_expr += '{}, '.format(val.name)

            dispatchfunc += call_expr
            body += '                 ' + dispatchfunc[:-2] + ');\n'
            body += '            }\n'
            body += '        }\n'
            body += '\n'
            body += '        VkCommandBuffer dispatch_rays_command_buffer = GetDispatchRaysCommandBuffer(commandBuffer);\n'
            body += '        if (dispatch_rays_command_buffer != VK_NULL_HANDLE)\n'
            body += '        {\n'

            dispatchfunc = 'func(dispatch_rays_command_buffer, ' + call_expr
            body += '             ' + dispatchfunc[:-2] + ');\n'

            body += '        }\n'
            body += '    }\n'
        else:
            override_call_expr = 'call_info, func, '
            for val in values:
                if val.is_pointer or val.is_array:
                    count = val.pointer_count
                    if self.is_handle(val.base_type) and val.base_type != 'VkCommandBuffer':
                        override_call_expr += '{}->GetPointer(), '.format(val.name)
                    else:
                        override_call_expr += '{}, '.format(val.name)
                else:
                    override_call_expr += '{}, '.format(val.name)

            override_call_expr = override_call_expr[:-2]
            body += '    if (IsRecording(commandBuffer))\n'
            body += '    {\n'
            body += '        {}({});\n'.format(self.DUMP_RESOURCES_OVERRIDES[name], override_call_expr)
            body += '    }\n'


        return body



#!/usr/bin/env python

"""Amalgamate json-cpp library sources into a single source and header file.

Works with python2.6+ and python3.4+.

Example of invocation (must be invoked from json-cpp top directory):
python amalgamate.py
"""
import os
import os.path
import sys

INCLUDE_PATH = "include/json"
SRC_PATH = "src/lib_json"

class AmalgamationFile:
    def __init__(self, top_dir):
        self.top_dir = top_dir
        self.blocks = []

    def add_text(self, text):
        if not text.endswith("\n"):
            text += "\n"
        self.blocks.append(text)

    def add_file(self, relative_input_path, wrap_in_comment=False):
        def add_marker(prefix):
            self.add_text("")
            self.add_text("// " + "/"*70)
            self.add_text("// %s of content of file: %s" % (prefix, relative_input_path.replace("\\","/")))
            self.add_text("// " + "/"*70)
            self.add_text("")
        add_marker("Beginning")
        f = open(os.path.join(self.top_dir, relative_input_path), "rt")
        content = f.read()
        if wrap_in_comment:
            content = "/*\n" + content + "\n*/"
        self.add_text(content)
        f.close()
        add_marker("End")
        self.add_text("\n\n\n\n")

    def get_value(self):
        return "".join(self.blocks).replace("\r\n","\n")

    def write_to(self, output_path):
        output_dir = os.path.dirname(output_path)
        if output_dir and not os.path.isdir(output_dir):
            os.makedirs(output_dir)
        f = open(output_path, "wb")
        f.write(str.encode(self.get_value(), 'UTF-8'))
        f.close()

def amalgamate_source(source_top_dir=None,
                       target_source_path=None,
                       header_include_path=None):
    """Produces amalgamated source.
       Parameters:
           source_top_dir: top-directory
           target_source_path: output .cpp path
           header_include_path: generated header path relative to target_source_path.
    """
    print("Amalgamating header...")
    header = AmalgamationFile(source_top_dir)
    header.add_text("/// Json-cpp amalgamated header (http://jsoncpp.sourceforge.net/).")
    header.add_text('/// It is intended to be used with #include "%s"' % header_include_path)
    header.add_file("LICENSE", wrap_in_comment=True)
    header.add_text("#ifndef JSON_AMALGAMATED_H_INCLUDED")
    header.add_text("# define JSON_AMALGAMATED_H_INCLUDED")
    header.add_text("/// If defined, indicates that the source file is amalgamated")
    header.add_text("/// to prevent private header inclusion.")
    header.add_text("#define JSON_IS_AMALGAMATION")
    header.add_file(os.path.join(INCLUDE_PATH, "version.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "allocator.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "config.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "forwards.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "json_features.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "value.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "reader.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "writer.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "assertions.h"))
    header.add_text("#endif //ifndef JSON_AMALGAMATED_H_INCLUDED")

    target_header_path = os.path.join(os.path.dirname(target_source_path), header_include_path)
    print("Writing amalgamated header to %r" % target_header_path)
    header.write_to(target_header_path)

    base, ext = os.path.splitext(header_include_path)
    forward_header_include_path = base + "-forwards" + ext
    print("Amalgamating forward header...")
    header = AmalgamationFile(source_top_dir)
    header.add_text("/// Json-cpp amalgamated forward header (http://jsoncpp.sourceforge.net/).")
    header.add_text('/// It is intended to be used with #include "%s"' % forward_header_include_path)
    header.add_text("/// This header provides forward declaration for all JsonCpp types.")
    header.add_file("LICENSE", wrap_in_comment=True)
    header.add_text("#ifndef JSON_FORWARD_AMALGAMATED_H_INCLUDED")
    header.add_text("# define JSON_FORWARD_AMALGAMATED_H_INCLUDED")
    header.add_text("/// If defined, indicates that the source file is amalgamated")
    header.add_text("/// to prevent private header inclusion.")
    header.add_text("#define JSON_IS_AMALGAMATION")
    header.add_file(os.path.join(INCLUDE_PATH, "version.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "allocator.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "config.h"))
    header.add_file(os.path.join(INCLUDE_PATH, "forwards.h"))
    header.add_text("#endif //ifndef JSON_FORWARD_AMALGAMATED_H_INCLUDED")

    target_forward_header_path = os.path.join(os.path.dirname(target_source_path),
                                               forward_header_include_path)
    print("Writing amalgamated forward header to %r" % target_forward_header_path)
    header.write_to(target_forward_header_path)

    print("Amalgamating source...")
    source = AmalgamationFile(source_top_dir)
    source.add_text("/// Json-cpp amalgamated source (http://jsoncpp.sourceforge.net/).")
    source.add_text('/// It is intended to be used with #include "%s"' % header_include_path)
    source.add_file("LICENSE", wrap_in_comment=True)
    source.add_text("")
    source.add_text('#include "%s"' % header_include_path)
    source.add_text("""
#ifndef JSON_IS_AMALGAMATION
#error "Compile with -I PATH_TO_JSON_DIRECTORY"
#endif
""")
    source.add_text("")
    source.add_file(os.path.join(SRC_PATH, "json_tool.h"))
    source.add_file(os.path.join(SRC_PATH, "json_reader.cpp"))
    source.add_file(os.path.join(SRC_PATH, "json_valueiterator.inl"))
    source.add_file(os.path.join(SRC_PATH, "json_value.cpp"))
    source.add_file(os.path.join(SRC_PATH, "json_writer.cpp"))

    print("Writing amalgamated source to %r" % target_source_path)
    source.write_to(target_source_path)

def main():
    usage = """%prog [options]
Generate a single amalgamated source and header file from the sources.
"""
    from optparse import OptionParser
    parser = OptionParser(usage=usage)
    parser.allow_interspersed_args = False
    parser.add_option("-s", "--source", dest="target_source_path", action="store", default="dist/jsoncpp.cpp",
        help="""Output .cpp source path. [Default: %default]""")
    parser.add_option("-i", "--include", dest="header_include_path", action="store", default="json/json.h",
        help="""Header include path. Used to include the header from the amalgamated source file. [Default: %default]""")
    parser.add_option("-t", "--top-dir", dest="top_dir", action="store", default=os.getcwd(),
        help="""Source top-directory. [Default: %default]""")
    parser.enable_interspersed_args()
    options, args = parser.parse_args()

    msg = amalgamate_source(source_top_dir=options.top_dir,
                             target_source_path=options.target_source_path,
                             header_include_path=options.header_include_path)
    if msg:
        sys.stderr.write(msg + "\n")
        sys.exit(1)
    else:
        print("Source successfully amalgamated")

if __name__ == "__main__":
    main()



#!/usr/bin/python3 -i
#
# Copyright (c) 2013-2024, The Khronos Group Inc.
#
# SPDX-License-Identifier: Apache-2.0

from generator import OutputGenerator, write


class DocIndexOutputGenerator(OutputGenerator):
    """DocIndexOutputGenerator - subclass of OutputGenerator.
    Generates an Index chapter for the spec.
    Similar to DocOutputGenerator, but writes a single file."""

    def apiName(self, name):
        """Return True if name is in the reserved API namespace.

        Delegates to the conventions object. """
        return self.genOpts.conventions.is_api_name(name)

    def beginFile(self, genOpts):
        OutputGenerator.beginFile(self, genOpts)
        #
        # Dictionaries are keyed by the name of the entity (e.g.
        # self.structs is keyed by structure names). Values are
        # text that will go after an index entry (e.g. to point to the FlagBits for a Flags type)
        # or None if there's no such data.
        self.basetypes = {}
        self.enums = {}
        self.flags = {}
        self.funcpointers = {}
        self.protos = {}
        self.structs = {}
        self.handles = {}
        self.defines = {}
        self.consts = {}

    def record_name(self, name_dict, name, extra_data=None):
        name_dict[name] = extra_data

    def output_name_dict(self, name_dict, title, prefix):

        anchor = title.lower().replace(' ', '-')
        write(f'[[index-{anchor}]]', file=self.outFile)

        write(f"### {title}", file=self.outFile)
        write('', file=self.outFile)

        for name in sorted(name_dict.keys()):
            text = f"* {prefix}{name}"
            extra_data = name_dict[name]
            if extra_data:
                text += f" -- {extra_data}"
            write(text, file=self.outFile)

        write('', file=self.outFile)

    def endFile(self):
        # Remove things from enums if they're a flag bits.
        flag_bits = set(self.enums.keys()).intersection(self.flags.values())
        for f in flag_bits:
            self.enums.pop(f)

        # This sets up the sub-chapter order
        pieces = (
            ('Base Types and Atoms', self.basetypes, 'basetype:'),
            ('Defines', self.defines, 'dlink:'),
            ('Enumerations', self.enums, 'elink:'),
            ('Flags and Flag Bits', self.flags, 'elink:'),
            ('Function Pointer Types', self.funcpointers, 'tlink:'),
            ('Functions', self.protos, 'flink:'),
            ('Handles', self.handles, 'slink:'),
            ('Structures', self.structs, 'slink:'),
        )
        write('[[index]]', file=self.outFile)
        write('## Index', file=self.outFile)
        write('', file=self.outFile)

        for title, name_dict, prefix in pieces:
            self.output_name_dict(name_dict, title, prefix)

        OutputGenerator.endFile(self)

    def genType(self, typeinfo, name, alias):
        """Generate type.

        Just adds to the right dict. In the case of bitmasks,
        tries to figure out the bit value type and store that in a dictionary instead."""
        OutputGenerator.genType(self, typeinfo, name, alias)
        typeElem = typeinfo.elem
        # If the type is a struct type, traverse the embedded <member> tags
        # generating a structure. Otherwise, emit the tag text.
        category = typeElem.get('category')
        if category in ('struct', 'union'):
            self.record_name(self.structs, name)
        elif category == 'bitmask':
            requiredEnum = typeElem.get('bitvalues')
            if requiredEnum is not None:
                self.record_name(self.flags, name, f"See also elink:{requiredEnum}")
        elif category == 'enum':
            self.record_name(self.enums, name)
        elif category == 'funcpointer':
            self.record_name(self.funcpointers, name)
        elif category == 'handle':
            self.record_name(self.handles, name)
        elif category == 'define':
            self.record_name(self.defines, name)
        elif category == 'basetype':
            # Don't add an entry for base types that are not API types
            # e.g. an API Bool type gets an entry, uint32_t does not
            if self.apiName(name):
                self.record_name(self.basetypes, name)

    def genGroup(self, groupinfo, groupName, alias):
        """Generate group (e.g. C "enum" type)."""
        OutputGenerator.genGroup(self, groupinfo, groupName, alias)

        if alias:
            return

        self.record_name(self.enums, groupName)

    def genEnum(self, enuminfo, name, alias):
        """Generate enumerant (compile time constants)."""
        OutputGenerator.genEnum(self, enuminfo, name, alias)

        if alias:
            return
        self.record_name(self.consts, name)

    def genCmd(self, cmdinfo, name, alias):
        """Generate command."""
        OutputGenerator.genCmd(self, cmdinfo, name, alias)

        if alias:
            return
        self.record_name(self.protos, name)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018 Valve Corporation
# Copyright (c) 2018 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write
from khronos_layer_func_table_generator import KhronosLayerFuncTableGenerator


class VulkanLayerFuncTableGeneratorOptions(VulkanBaseGeneratorOptions):
    """Eliminates JSON black_lists and platform_types files, which are not necessary for
    function table generation.
    Options for Vulkan layer function table C++ code generation.
    """

    def __init__(
        self,
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            None,
            None,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )
        self.begin_end_file_data.specific_headers.extend((
            'encode/custom_vulkan_api_call_encoders.h',
            'generated/generated_vulkan_api_call_encoders.h',
            'layer/trace_layer.h',
            'util/defines.h',
        ))
        self.begin_end_file_data.system_headers.append('unordered_map')
        self.begin_end_file_data.namespaces.append('gfxrecon')

class VulkanLayerFuncTableGenerator(VulkanBaseGenerator, KhronosLayerFuncTableGenerator):
    """LayerFuncTableGenerator - subclass of VulkanBaseGenerator.
    Generates C++ function table for the Vulkan API calls exported by the layer.
    Generate Vulkan layer function table C++ type declarations.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

        # The trace layer does not currently implement or export the instance version query
        self.APICALL_BLACKLIST = ['vkEnumerateInstanceVersion']

        # These functions are provided directly by the layer, and are not encoded
        self.LAYER_FUNCTIONS = [
            'vkGetInstanceProcAddr', 'vkGetDeviceProcAddr',
            'vkEnumerateInstanceLayerProperties',
            'vkEnumerateDeviceLayerProperties',
            'vkEnumerateInstanceExtensionProperties',
            'vkEnumerateDeviceExtensionProperties'
        ]

    def endFile(self):
        """Method override."""

        KhronosLayerFuncTableGenerator.write_layer_func_table_contents(self, self.LAYER_FUNCTIONS, 100)
        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def write_custom_layer_func_table_contents(self, api_data, align_col):
        """ Method override """
        # Manually output the physical device proc address function as its name doesn't
        # match the scheme used by skip_func_list:
        align = align_col - len('vk_layerGetPhysicalDeviceProcAddr')
        write('    { "vk_layerGetPhysicalDeviceProcAddr",%sreinterpret_cast<PFN_vkVoidFunction>(vulkan_entry::GetPhysicalDeviceProcAddr) },' % (' ' * align), file=self.outFile)


    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_cmd_params:
            return True
        return False



#!/usr/bin/python3 -i
#
# Copyright (c) 2019 Valve Corporation
# Copyright (c) 2019-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write
from khronos_struct_handle_wrappers_body_generator import KhronosStructHandleWrappersBodyGenerator


class VulkanStructHandleWrappersBodyGeneratorOptions(VulkanBaseGeneratorOptions):
    """Options for generating functions to wrap Vulkan struct member handles at API capture."""

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'generated/generated_vulkan_struct_handle_wrappers.h',
            '',
            'vulkan/vk_layer.h',
        ))
        self.begin_end_file_data.namespaces.extend((
            'gfxrecon',
            'encode',
            'vulkan_wrappers',
        ))
        self.begin_end_file_data.common_api_headers = []

class VulkanStructHandleWrappersBodyGenerator(VulkanBaseGenerator, KhronosStructHandleWrappersBodyGenerator):
    """VulkanStructHandleWrappersBodyGenerator - subclass of VulkanBaseGenerator.
    Generates C++ functions responsible for wrapping struct member handles
    when recording Vulkan API call parameter data.
    Generate C++ functions for Vulkan struct member handle wrapping at API capture.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def endFile(self):
        """Method override."""
        KhronosStructHandleWrappersBodyGenerator.write_struct_handle_wrapper_content(self)
        self.newline()

        # Finish processing in superclass
        VulkanBaseGenerator.endFile(self)

    def need_feature_generation(self):
        """Indicates that the current feature has C++ code to generate."""
        if self.feature_struct_members:
            return True
        return False

    def write_special_case_struct_handling(self):
        """Method override."""
        write(
            '    case VK_STRUCTURE_TYPE_LOADER_INSTANCE_CREATE_INFO:',
            file=self.outFile
        )
        write(
            '        copy = reinterpret_cast<VkBaseInStructure*>(MakeUnwrapStructs(reinterpret_cast<const VkLayerInstanceCreateInfo*>(base), 1, unwrap_memory));',
            file=self.outFile
        )
        write('        break;', file=self.outFile)
        write(
            '    case VK_STRUCTURE_TYPE_LOADER_DEVICE_CREATE_INFO:',
            file=self.outFile
        )
        write(
            '        copy = reinterpret_cast<VkBaseInStructure*>(MakeUnwrapStructs(reinterpret_cast<const VkLayerDeviceCreateInfo*>(base), 1, unwrap_memory));',
            file=self.outFile
        )
        write('        break;', file=self.outFile)

    def has_special_case_handle_unwrapping(self, name):
        """Method override."""
        if (name == 'VkGeneratedCommandsMemoryRequirementsInfoEXT'):
            return True
        return False

    def get_special_case_handle_wrapping(self, name):
        """Method override."""
        # Workaround for spec missing const in VkGeneratedCommandsMemoryRequirementsInfoEXT::pNext, should be removed next header update
        if (name == 'VkGeneratedCommandsMemoryRequirementsInfoEXT'):
            return '            value->pNext = const_cast<void*>(UnwrapPNextStructHandles(value->pNext, unwrap_memory));\n'
        return ''



# Copyright 2007 Baptiste Lepilleur and The JsonCpp Authors
# Distributed under MIT license, or public domain if desired and
# recognized in your jurisdiction.
# See file LICENSE for detail or copy at http://jsoncpp.sourceforge.net/LICENSE

from __future__ import print_function
import glob
import os.path
for path in glob.glob('*.json'):
    text = file(path,'rt').read()
    target = os.path.splitext(path)[0] + '.expected'
    if os.path.exists(target):
        print('skipping:', target)
    else:
        print('creating:', target)
        file(target,'wt').write(text)




#!/usr/bin/python3 -i
#
# Copyright (c) 2022-2024 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#

from khronos_base_generator import write

class KhronosStructToJsonHeaderGenerator():
    """KhronosStructToJsonHeaderGenerator
    Generate C++ function declarations to serialize Khronos structures to JSON"""

    def should_decode_struct(self, struct):
        """Method indended to be overridden.
        Indicates that the provided struct is a struct we want to decode"""
        return True

    def write_parent_child_to_json_func(self):
        struct_type = self.get_struct_type_enum_name()
        body = '\n'
        body += 'template <typename T>\n'
        body += 'void ParentChildFieldToJson(nlohmann::ordered_json& jdata, const T* data, const util::JsonOptions& options = util::JsonOptions())\n'
        body += '{\n'
        body += '    // First read in the type to know which child we need to handle\n'
        body += f'    {struct_type} struct_type;\n'
        body += '    FieldToJson(jdata["type"], struct_type, options);\n'
        body += '\n'
        body += '    switch (struct_type)\n'
        body += '    {\n'
        body += '        default:\n'
        body += '        {\n'
        body += '            GFXRECON_LOG_WARNING("ParentChildFieldToJson: unrecognized child structure type %d", struct_type);\n'
        body += '            break;\n'
        body += '        }\n'
        for child_list in self.children_structs:
            for child in self.children_structs[child_list]:
                struct_type_name = self.struct_type_names[child]
                body += f'        case {struct_type_name}:\n'
                body += '        {\n'
                body += f'            FieldToJson(jdata, reinterpret_cast<const Decoded_{child}*>(data), options);\n'
                body += '            break;\n'
                body += '        }\n'
        body += '    }\n'
        body += '}\n'
        write(body, file=self.outFile)

    def write_header_contents(self):
        for struct in self.get_all_filtered_struct_names():
            if self.should_decode_struct(struct):
                body = "void FieldToJson(nlohmann::ordered_json& jdata, const Decoded_{0}* data, const util::JsonOptions& options = util::JsonOptions());".format(struct)
                write(body, file=self.outFile)
        self.newline()

        # If there are parent/child structs, write the utility function to help
        if len(self.children_structs) > 0:
            self.write_parent_child_to_json_func()

        var_name = self.get_extended_struct_var_name()
        prefix = self.get_extended_struct_node_prefix()
        if len(var_name) > 0 and len(prefix) > 0:
            write(f'/// Works out the type of the struct at the end of a {var_name} pointer and dispatches', file=self.outFile)
            write('/// recursively to the FieldToJson for that.', file=self.outFile)
            write(f'void FieldToJson(nlohmann::ordered_json& jdata, const {prefix}Node* data, const util::JsonOptions& options = util::JsonOptions());', file=self.outFile)



#!/usr/bin/env python3
# Copyright (c) 2019-2024, The Khronos Group Inc.
# SPDX-License-Identifier: Apache-2.0

from itertools import product

from shared import (PLATFORMS, TRUE_FALSE, VS_VERSION, make_win_artifact_name,
                    output_json)

if __name__ == "__main__":

    configs = {}
    for  platform, debug, uwp in product(PLATFORMS, (False,), TRUE_FALSE):
        # No need to support ARM/ARM64 except for UWP.
        if not uwp and (platform.lower() == 'arm' or platform.lower() == 'arm64'):
            continue

        label = [platform]
        config = []
        generator = VS_VERSION
        config.append('-A ' + platform)
        config.append('-DDYNAMIC_LOADER=ON')
        if debug:
            label.append('debug')
        if uwp:
            label.append('UWP')
            config.append('-DCMAKE_SYSTEM_NAME=WindowsStore -DCMAKE_SYSTEM_VERSION=10.0')
        name = '_'.join(label)
        configs[name] = {
            'generator': generator,
            'buildType': 'Debug' if debug else 'RelWithDebInfo',
            'cmakeArgs': ' '.join(config)
        }
        if not debug:
            configs[name]['artifactName'] = make_win_artifact_name(
                platform, uwp)

    output_json(configs)



#!/usr/bin/python3 -i
#
# Copyright (c) 2018-2021 Valve Corporation
# Copyright (c) 2018-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.
#
# The content of this file was derived from the Khronos Registry cgenerator.py
# and related Python files found in the KhronosGroup/Vulkan-Headers GitHub repository.
#
# Copyright (c) 2013-2016 The Khronos Group Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# The content of this file was derived from the Khronos Registry cgenerator.py
# and related Python files found in the KhronosGroup/Vulkan-Headers GitHub repository.

import os,re,sys,json
from collections import OrderedDict
from khronos_base_generator import (KhronosBaseGeneratorOptions, KhronosBaseGenerator, make_re_string, ValueInfo, write)
from generator import GeneratorOptions, OutputGenerator, noneStr, regSortFeatures, write
from xrconventions import OpenXRConventions

# Descriptive names for various regexp patterns used to select versions and extensions.
_default_extensions = 'openxr'
_extensions = _features = []
_emit_extensions = []

# Exclude extensions from code generation.
# Note this doesn't totally eliminate them. ToString and ToJson functions for enums
# will always be generated but functions and structs can be screened out by editing
# the blocklists files such as the vulkan default openxr default "openxr_generators/blacklists.json".
# Note, this doesn't hide them from the  application, but lets them bypass our
# layer during capture, meaning we will not call any of their functions at
# replay.
# To screen an extension out from the list reported to the application it should
# be added to the list kUnsupportedDeviceExtensions in trace_layer.cpp.
_remove_extensions = []

# Turn lists of names/patterns into matching regular expressions.
# From Khronos genvk.py
_add_extensions_pat = make_re_string(_extensions)
_remove_extensions_pat = make_re_string(_remove_extensions)
_emit_extensions_pat = make_re_string(_emit_extensions, '.*')
_features_pat = make_re_string(_features, '.*')


class OpenXrBaseGeneratorOptions(KhronosBaseGeneratorOptions):
    """OpenXrBaseGeneratorOptions - subclass of KhronosGeneratorOptions.
    Options for OpenXr API parameter encoding and decoding C++ code generation.

    Adds options used by FrameworkGenerator objects during C++ language
    code generation.

    Additional members
      blacklists - Path to JSON file listing apicalls and structs to ignore.
      platform_types - Path to JSON file listing platform (WIN32, X11, etc.)
        specific types that are defined outside of the OpenXr header.

    Additional members (from Khronos Registry COptionsGenerator)
      prefix_text - list of strings to prefix generated header with
        (usually a copyright statement + calling convention macros).
      protect_file - True if multiple inclusion protection should be
        generated (based on the filename) around the entire header.
      apicall - string to use for the function declaration prefix,
        such as APICALL on Windows.
      apientry - string to use for the calling convention macro,
        in typedefs, such as APIENTRY.
      apientryp - string to use for the calling convention macro
        in function pointer typedefs, such as APIENTRYP.
      indent_func_proto - True if prototype declarations should put each
        parameter on a separate line
      indent_func_pointer - True if typedefed function pointers should put each
        parameter on a separate line
      align_func_param - if nonzero and parameters are being put on a
        separate line, align parameter names at the specified column
      replay_overrides - Path to JSON file listing OpenXr API calls to
        override on replay.
      dump_resources_overrides - Path to JSON file listing OpenXr API
        calls to override on replay.
      replay_async_overrides - Path to JSON file listing OpenXr API calls
        to override on replay.
      extra_headers - headers to include in addition to the standard Khronos API
    """

    def __init__(
        self,
        blacklists=None,
        platform_types=None,
        # Khronos CGeneratorOptions
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        conventions=OpenXRConventions(),
        apicall='XRAPI_ATTR ',
        apientry='XRAPI_CALL ',
        apientryp='XRAPI_PTR *',
        indent_func_proto=True,
        align_func_param=48,
        sort_procedure=regSortFeatures,
        apiname='openxr',
        profile=None,
        versions=_features_pat,
        emit_versions=_features_pat,
        default_extensions=_default_extensions,
        add_extensions=_add_extensions_pat,
        remove_extensions=_remove_extensions_pat,
        emit_extensions=_emit_extensions_pat,
        replay_overrides=None,
        dump_resources_overrides=None,
        replay_async_overrides=None,
        extra_headers=[],
        extra_manual_commands=[],
    ):
        KhronosBaseGeneratorOptions.__init__(
            self,
            blacklists=blacklists,
            platform_types=platform_types,
            filename=filename,
            directory=directory,
            prefix_text=prefix_text,
            protect_file=protect_file,
            protect_feature=protect_feature,
            conventions=conventions,
            apicall=apicall,
            apientry=apientry,
            apientryp=apientryp,
            indent_func_proto=True,
            align_func_param=align_func_param,
            sort_procedure=sort_procedure,
            apiname=apiname,
            profile=profile,
            versions=versions,
            emit_versions=emit_versions,
            default_extensions=_default_extensions,
            add_extensions=add_extensions,
            remove_extensions=remove_extensions,
            emit_extensions=emit_extensions,
            replay_overrides=replay_overrides,
            dump_resources_overrides=dump_resources_overrides,
            replay_async_overrides=replay_async_overrides,
            extra_headers=extra_headers,
            extra_manual_commands=extra_manual_commands,
        )

        self.begin_end_file_data.guards.append(('if', 'ENABLE_OPENXR_SUPPORT'))
        self.begin_end_file_data.common_api_headers = [
            'format/platform_types.h',
            '',
            'openxr/openxr.h',
            'openxr/openxr_loader_negotiation.h',
            'openxr/openxr_platform.h',
        ]

class OpenXrBaseGenerator(KhronosBaseGenerator):
    """OpenXrBaseGenerator - subclass of KhronosBaseGenerator.
    Base class providing common operations used to generate C++-language code for framework
      components that encode and decode OpenXr API parameters.
    Base class for OpenXr API parameter encoding and decoding generators.
    """

    def __init__(
        self,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        KhronosBaseGenerator.__init__(
            self,
            err_file = err_file,
            warn_file = warn_file,
            diag_file = diag_file)

        self.SUPPORTED_SUBSETS= [ "openxr"]

        # Platform specific structure types that have been defined extarnally to the OpenXr header.
        self.PLATFORM_STRUCTS.append('timespec')

        # Add OpenXr specific encode_types
        self.encode_types['XR_DEFINE_ATOM'] = 'UInt64'
        self.encode_types['XR_DEFINE_OPAQUE_64'] = 'UInt64'

        # Add Vulkan handle types to handle list
        self.handle_names.add('VkInstance')
        self.handle_names.add('VkPhysicalDevice')
        self.handle_names.add('VkDevice')
        self.handle_names.add('VkImage')
        self.handle_names.add('VkCommandBuffer')
        self.handle_names.add('VkQueue')
        self.handle_names.add('VkSurfaceKHR')
        self.handle_names.add('VkSwapchainKHR')

        # Add Vulkan enums
        self.enum_names.add('VkResult')
        self.enum_names.add('VkFormat')
        self.enum_names.add('VkFilter')
        self.enum_names.add('VkSamplerMipmapMode')
        self.enum_names.add('VkSamplerAddressMode')
        self.enum_names.add('VkComponentSwizzle')

        # Vulkan flag types
        # Updated as of v1.4.304 using
        # TODO: Automate extraction of this information from the Vulkan xml or header
        #
        # sed -n "s/typedef \(VkFlags6*4*\)  *\(.*\);/'\2': '\1',/p" vulkan_core.h
        vulkan_flags_types = {
            'VkAccessFlags': 'VkFlags',
            'VkImageAspectFlags': 'VkFlags',
            'VkFormatFeatureFlags': 'VkFlags',
            'VkImageCreateFlags': 'VkFlags',
            'VkSampleCountFlags': 'VkFlags',
            'VkImageUsageFlags': 'VkFlags',
            'VkInstanceCreateFlags': 'VkFlags',
            'VkMemoryHeapFlags': 'VkFlags',
            'VkMemoryPropertyFlags': 'VkFlags',
            'VkQueueFlags': 'VkFlags',
            'VkDeviceCreateFlags': 'VkFlags',
            'VkDeviceQueueCreateFlags': 'VkFlags',
            'VkPipelineStageFlags': 'VkFlags',
            'VkMemoryMapFlags': 'VkFlags',
            'VkSparseMemoryBindFlags': 'VkFlags',
            'VkSparseImageFormatFlags': 'VkFlags',
            'VkFenceCreateFlags': 'VkFlags',
            'VkSemaphoreCreateFlags': 'VkFlags',
            'VkEventCreateFlags': 'VkFlags',
            'VkQueryPipelineStatisticFlags': 'VkFlags',
            'VkQueryPoolCreateFlags': 'VkFlags',
            'VkQueryResultFlags': 'VkFlags',
            'VkBufferCreateFlags': 'VkFlags',
            'VkBufferUsageFlags': 'VkFlags',
            'VkBufferViewCreateFlags': 'VkFlags',
            'VkImageViewCreateFlags': 'VkFlags',
            'VkShaderModuleCreateFlags': 'VkFlags',
            'VkPipelineCacheCreateFlags': 'VkFlags',
            'VkColorComponentFlags': 'VkFlags',
            'VkPipelineCreateFlags': 'VkFlags',
            'VkPipelineShaderStageCreateFlags': 'VkFlags',
            'VkCullModeFlags': 'VkFlags',
            'VkPipelineVertexInputStateCreateFlags': 'VkFlags',
            'VkPipelineInputAssemblyStateCreateFlags': 'VkFlags',
            'VkPipelineTessellationStateCreateFlags': 'VkFlags',
            'VkPipelineViewportStateCreateFlags': 'VkFlags',
            'VkPipelineRasterizationStateCreateFlags': 'VkFlags',
            'VkPipelineMultisampleStateCreateFlags': 'VkFlags',
            'VkPipelineDepthStencilStateCreateFlags': 'VkFlags',
            'VkPipelineColorBlendStateCreateFlags': 'VkFlags',
            'VkPipelineDynamicStateCreateFlags': 'VkFlags',
            'VkPipelineLayoutCreateFlags': 'VkFlags',
            'VkShaderStageFlags': 'VkFlags',
            'VkSamplerCreateFlags': 'VkFlags',
            'VkDescriptorPoolCreateFlags': 'VkFlags',
            'VkDescriptorPoolResetFlags': 'VkFlags',
            'VkDescriptorSetLayoutCreateFlags': 'VkFlags',
            'VkAttachmentDescriptionFlags': 'VkFlags',
            'VkDependencyFlags': 'VkFlags',
            'VkFramebufferCreateFlags': 'VkFlags',
            'VkRenderPassCreateFlags': 'VkFlags',
            'VkSubpassDescriptionFlags': 'VkFlags',
            'VkCommandPoolCreateFlags': 'VkFlags',
            'VkCommandPoolResetFlags': 'VkFlags',
            'VkCommandBufferUsageFlags': 'VkFlags',
            'VkQueryControlFlags': 'VkFlags',
            'VkCommandBufferResetFlags': 'VkFlags',
            'VkStencilFaceFlags': 'VkFlags',
            'VkSubgroupFeatureFlags': 'VkFlags',
            'VkPeerMemoryFeatureFlags': 'VkFlags',
            'VkMemoryAllocateFlags': 'VkFlags',
            'VkCommandPoolTrimFlags': 'VkFlags',
            'VkDescriptorUpdateTemplateCreateFlags': 'VkFlags',
            'VkExternalMemoryHandleTypeFlags': 'VkFlags',
            'VkExternalMemoryFeatureFlags': 'VkFlags',
            'VkExternalFenceHandleTypeFlags': 'VkFlags',
            'VkExternalFenceFeatureFlags': 'VkFlags',
            'VkFenceImportFlags': 'VkFlags',
            'VkSemaphoreImportFlags': 'VkFlags',
            'VkExternalSemaphoreHandleTypeFlags': 'VkFlags',
            'VkExternalSemaphoreFeatureFlags': 'VkFlags',
            'VkResolveModeFlags': 'VkFlags',
            'VkDescriptorBindingFlags': 'VkFlags',
            'VkSemaphoreWaitFlags': 'VkFlags',
            'VkPipelineCreationFeedbackFlags': 'VkFlags',
            'VkToolPurposeFlags': 'VkFlags',
            'VkPrivateDataSlotCreateFlags': 'VkFlags',
            'VkPipelineStageFlags2': 'VkFlags64',
            'VkPipelineStageFlagBits2': 'VkFlags64',
            'VkAccessFlags2': 'VkFlags64',
            'VkAccessFlagBits2': 'VkFlags64',
            'VkSubmitFlags': 'VkFlags',
            'VkRenderingFlags': 'VkFlags',
            'VkFormatFeatureFlags2': 'VkFlags64',
            'VkFormatFeatureFlagBits2': 'VkFlags64',
            'VkMemoryUnmapFlags': 'VkFlags',
            'VkPipelineCreateFlags2': 'VkFlags64',
            'VkPipelineCreateFlagBits2': 'VkFlags64',
            'VkBufferUsageFlags2': 'VkFlags64',
            'VkBufferUsageFlagBits2': 'VkFlags64',
            'VkHostImageCopyFlags': 'VkFlags',
            'VkCompositeAlphaFlagsKHR': 'VkFlags',
            'VkSurfaceTransformFlagsKHR': 'VkFlags',
            'VkSwapchainCreateFlagsKHR': 'VkFlags',
            'VkDeviceGroupPresentModeFlagsKHR': 'VkFlags',
            'VkDisplayModeCreateFlagsKHR': 'VkFlags',
            'VkDisplayPlaneAlphaFlagsKHR': 'VkFlags',
            'VkDisplaySurfaceCreateFlagsKHR': 'VkFlags',
            'VkVideoCodecOperationFlagsKHR': 'VkFlags',
            'VkVideoChromaSubsamplingFlagsKHR': 'VkFlags',
            'VkVideoComponentBitDepthFlagsKHR': 'VkFlags',
            'VkVideoCapabilityFlagsKHR': 'VkFlags',
            'VkVideoSessionCreateFlagsKHR': 'VkFlags',
            'VkVideoSessionParametersCreateFlagsKHR': 'VkFlags',
            'VkVideoBeginCodingFlagsKHR': 'VkFlags',
            'VkVideoEndCodingFlagsKHR': 'VkFlags',
            'VkVideoCodingControlFlagsKHR': 'VkFlags',
            'VkVideoDecodeCapabilityFlagsKHR': 'VkFlags',
            'VkVideoDecodeUsageFlagsKHR': 'VkFlags',
            'VkVideoDecodeFlagsKHR': 'VkFlags',
            'VkVideoEncodeH264CapabilityFlagsKHR': 'VkFlags',
            'VkVideoEncodeH264StdFlagsKHR': 'VkFlags',
            'VkVideoEncodeH264RateControlFlagsKHR': 'VkFlags',
            'VkVideoEncodeH265CapabilityFlagsKHR': 'VkFlags',
            'VkVideoEncodeH265StdFlagsKHR': 'VkFlags',
            'VkVideoEncodeH265CtbSizeFlagsKHR': 'VkFlags',
            'VkVideoEncodeH265TransformBlockSizeFlagsKHR': 'VkFlags',
            'VkVideoEncodeH265RateControlFlagsKHR': 'VkFlags',
            'VkVideoDecodeH264PictureLayoutFlagsKHR': 'VkFlags',
            'VkPerformanceCounterDescriptionFlagsKHR': 'VkFlags',
            'VkAcquireProfilingLockFlagsKHR': 'VkFlags',
            'VkVideoEncodeFlagsKHR': 'VkFlags',
            'VkVideoEncodeCapabilityFlagsKHR': 'VkFlags',
            'VkVideoEncodeRateControlModeFlagsKHR': 'VkFlags',
            'VkVideoEncodeFeedbackFlagsKHR': 'VkFlags',
            'VkVideoEncodeUsageFlagsKHR': 'VkFlags',
            'VkVideoEncodeContentFlagsKHR': 'VkFlags',
            'VkVideoEncodeRateControlFlagsKHR': 'VkFlags',
            'VkVideoEncodeAV1CapabilityFlagsKHR': 'VkFlags',
            'VkVideoEncodeAV1StdFlagsKHR': 'VkFlags',
            'VkVideoEncodeAV1SuperblockSizeFlagsKHR': 'VkFlags',
            'VkVideoEncodeAV1RateControlFlagsKHR': 'VkFlags',
            'VkDebugReportFlagsEXT': 'VkFlags',
            'VkPipelineRasterizationStateStreamCreateFlagsEXT': 'VkFlags',
            'VkExternalMemoryHandleTypeFlagsNV': 'VkFlags',
            'VkExternalMemoryFeatureFlagsNV': 'VkFlags',
            'VkConditionalRenderingFlagsEXT': 'VkFlags',
            'VkSurfaceCounterFlagsEXT': 'VkFlags',
            'VkPipelineViewportSwizzleStateCreateFlagsNV': 'VkFlags',
            'VkPipelineDiscardRectangleStateCreateFlagsEXT': 'VkFlags',
            'VkPipelineRasterizationConservativeStateCreateFlagsEXT': 'VkFlags',
            'VkPipelineRasterizationDepthClipStateCreateFlagsEXT': 'VkFlags',
            'VkDebugUtilsMessengerCallbackDataFlagsEXT': 'VkFlags',
            'VkDebugUtilsMessageTypeFlagsEXT': 'VkFlags',
            'VkDebugUtilsMessageSeverityFlagsEXT': 'VkFlags',
            'VkDebugUtilsMessengerCreateFlagsEXT': 'VkFlags',
            'VkPipelineCoverageToColorStateCreateFlagsNV': 'VkFlags',
            'VkPipelineCoverageModulationStateCreateFlagsNV': 'VkFlags',
            'VkValidationCacheCreateFlagsEXT': 'VkFlags',
            'VkGeometryFlagsKHR': 'VkFlags',
            'VkGeometryInstanceFlagsKHR': 'VkFlags',
            'VkBuildAccelerationStructureFlagsKHR': 'VkFlags',
            'VkPipelineCompilerControlFlagsAMD': 'VkFlags',
            'VkShaderCorePropertiesFlagsAMD': 'VkFlags',
            'VkPipelineCoverageReductionStateCreateFlagsNV': 'VkFlags',
            'VkHeadlessSurfaceCreateFlagsEXT': 'VkFlags',
            'VkPresentScalingFlagsEXT': 'VkFlags',
            'VkPresentGravityFlagsEXT': 'VkFlags',
            'VkIndirectStateFlagsNV': 'VkFlags',
            'VkIndirectCommandsLayoutUsageFlagsNV': 'VkFlags',
            'VkDeviceMemoryReportFlagsEXT': 'VkFlags',
            'VkDeviceDiagnosticsConfigFlagsNV': 'VkFlags',
            'VkGraphicsPipelineLibraryFlagsEXT': 'VkFlags',
            'VkAccelerationStructureMotionInfoFlagsNV': 'VkFlags',
            'VkAccelerationStructureMotionInstanceFlagsNV': 'VkFlags',
            'VkImageCompressionFlagsEXT': 'VkFlags',
            'VkImageCompressionFixedRateFlagsEXT': 'VkFlags',
            'VkDeviceAddressBindingFlagsEXT': 'VkFlags',
            'VkFrameBoundaryFlagsEXT': 'VkFlags',
            'VkBuildMicromapFlagsEXT': 'VkFlags',
            'VkMicromapCreateFlagsEXT': 'VkFlags',
            'VkPhysicalDeviceSchedulingControlsFlagsARM': 'VkFlags64',
            'VkPhysicalDeviceSchedulingControlsFlagBitsARM': 'VkFlags64',
            'VkMemoryDecompressionMethodFlagBitsNV': 'VkFlags64',
            'VkMemoryDecompressionMethodFlagsNV': 'VkFlags64',
            'VkDirectDriverLoadingFlagsLUNARG': 'VkFlags',
            'VkOpticalFlowGridSizeFlagsNV': 'VkFlags',
            'VkOpticalFlowUsageFlagsNV': 'VkFlags',
            'VkOpticalFlowSessionCreateFlagsNV': 'VkFlags',
            'VkOpticalFlowExecuteFlagsNV': 'VkFlags',
            'VkShaderCreateFlagsEXT': 'VkFlags',
            'VkIndirectCommandsInputModeFlagsEXT': 'VkFlags',
            'VkIndirectCommandsLayoutUsageFlagsEXT': 'VkFlags',
            'VkAccelerationStructureCreateFlagsKHR': 'VkFlags',
        }
        self.flags_types.update(vulkan_flags_types)

    def beginFile(self, gen_opts):
        """Method override."""
        KhronosBaseGenerator.beginFile(self, gen_opts)

    def endFile(self):
        """Method override. Generate code for the feature."""
        KhronosBaseGenerator.endFile(self)

    def beginFeature(self, interface, emit):
        """Method override. Start processing in superclass."""
        KhronosBaseGenerator.beginFeature(self, interface, emit)

    def endFeature(self):
        """Method override. Generate code for the feature."""
        KhronosBaseGenerator.endFeature(self)

    def genStruct(self, typeinfo, typename, alias):
        """Method override."""
        super().genStruct(typeinfo, typename, alias)

        if not alias:
            self.check_struct_member_handles(
                typename, self.structs_with_handles
            )

    def get_feature_protect(self, interface):
        """Return appropriate feature protect string from 'platform' tag on feature.
        """
        # TODO: This should probably be in a JSON file.
        platform_dict = {
             'android': 'XR_USE_PLATFORM_ANDROID',
            'wayland': 'XR_USE_PLATFORM_WAYLAND',
            'win32': 'XR_USE_PLATFORM_WIN32',
            'xcb': 'XR_USE_PLATFORM_XCB',
            'xlib': 'XR_USE_PLATFORM_XLIB',
        }

        platform = interface.get('platform')
        if platform and platform in platform_dict:
            return platform_dict[platform]
        return None

    def get_handle_wrapper(self, base_type):
        handle_wrapper = ''
        if base_type.startswith('Vk'):
            handle_wrapper = 'vulkan_wrappers::'
        else:
            handle_wrapper = 'openxr_wrappers::'
        handle_wrapper += base_type[2:] + 'Wrapper'
        return handle_wrapper

    def get_atom_encode_text(self, type_name):
        return 'OpenXrAtom'

    def get_opaque_encode_text(self, type_name):
        return 'OpenXrOpaque'

    def add_remaining_base_types(self, name, type_info):
            if type_info == 'XR_DEFINE_ATOM':
                self.atom_names.add(name)
            elif type_info == 'XR_DEFINE_OPAQUE_64':
                self.opaque_names.add(name)
            elif not name == 'XrTime':
                super().add_remaining_base_types(name, type_info)



#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

from CppHeaderParser import CppHeader, CppHeaderParser

SAL_TOKENS = [
    '__in_ecount', '__in_ecount_opt', '__RPC__deref_out',
    '__RPC__deref_out_opt', '__RPC__in', '_Always_', '_COM_Outptr_',
    '_COM_Outptr_opt_', '_COM_Outptr_opt_result_maybenull_',
    '_Field_size_bytes_full_', '_Field_size_bytes_full_opt_',
    '_Field_size_full_', '_Field_size_full_opt_', '_Field_z_', '_In_',
    '_In_opt_', '_In_range_', '_In_reads_', '_In_reads_bytes_',
    '_In_reads_bytes_opt_', '_In_reads_opt_', '_In_z_', '_Inout_',
    '_Inout_opt_', '_Inout_updates_bytes_', '_Out_', '_Out_opt_',
    '_Out_writes_', '_Out_writes_all_', '_Out_writes_all_opt_',
    '_Out_writes_bytes_', '_Out_writes_bytes_opt_', '_Out_writes_bytes_to_',
    '_Out_writes_opt_', '_Out_writes_to_opt_',
    '_Outptr_opt_result_bytebuffer_', '_Field_size_', '_In_opt_count_',
    '_In_count_'
]

original_warning_print = CppHeaderParser.warning_print


def dx12_warning_print(fmt, *args):
    """Override CppHeaderParser.warning_print."""
    # Silence warnings for unresolved SAL tokens.
    if (not fmt.startswith('WARN unresolved')) or (args[0] not in SAL_TOKENS):
        original_warning_print(fmt, *args)


CppHeaderParser.warning_print = dx12_warning_print

original_is_method_namestack = CppHeaderParser.is_method_namestack
original_is_property_namestack = CppHeaderParser.is_property_namestack


def dx12_is_method_namestack(stack):
    """Function override."""
    r = original_is_method_namestack(stack)
    # Additional check for D3D12 headers.
    if r:
        index = stack.index("(")
        # 0: return type, 1: function name
        # if "(" index is less than 2, this is a property, not a method
        if index < 2:
            r = False
    return r


def dx12_is_property_namestack(name_stack):
    """Function override."""
    r = original_is_property_namestack(name_stack)
    if not r:
        if "(" in name_stack and ")" in name_stack:
            index = name_stack.index("(")
            # if "(" index is less than 2, this is a property, not a method
            if index < 2:
                r = True
    return r


CppHeaderParser.is_method_namestack = dx12_is_method_namestack
CppHeaderParser.is_property_namestack = dx12_is_property_namestack


class Dx12CppClass():
    """This struct is simliar to CppHeaderParser.CppClass. In order to add data into CppHeader manually."""

    def __init__(self):
        self.classes = {}
        self.functions = []
        self.enums = []


class Dx12CppHeader(CppHeader):

    def __init__(self, header_file_name, encoding=None, **kwargs):
        """Method override.
           Custom CppHeader implementation to modify the content of the
           header file to remove D3D12 specific syntax before parsing
           with the CppHeader base class.
        """
        source = ''
        with open(header_file_name, 'r') as fd:
            source = self.preprocess_file(fd.readlines())
        CppHeader.__init__(self, source, "string", encoding, **kwargs)

    def preprocess_file(self, lines):
        """Preprocess header file to remove MIDL macros and CINTERACE declarations."""
        source = ''
        interface_scope = 0
        ignore_lines = False
        enum_scope = False
        retval_param = False

        for line in lines:

            if 'DEFINE_ENUM_FLAG_OPERATORS(' in line:
                continue
            
            if interface_scope == 0:
                source += line
                enum_scope = line.startswith('typedef enum ')
                if (line.startswith('#if') and ('!defined(CINTERFACE)' in line)) or enum_scope:
                    interface_scope = 1
            else:
                if enum_scope:
                    if line.find(', *') != -1:
                        interface_scope = 0
                        ignore_lines = False
                else:
                    if line.startswith('#if'):
                        interface_scope += 1
                        
                        if 'defined(_MSC_VER) || !defined(_WIN32)' in line:
                            retval_param = True

                    elif line.startswith('#else'):
                        if interface_scope == 1:
                            ignore_lines = True

                        if retval_param is True:
                            ignore_lines = True

                    elif line.startswith('#endif'):
                        interface_scope -= 1

                        if retval_param == True:
                            ignore_lines = False

                        retval_param = False

                        if interface_scope == 0:
                            ignore_lines = False

                if not ignore_lines:
                    new_line = line
                    if enum_scope:
                        new_line = new_line.replace('#if', '//#if')
                        new_line = new_line.replace('#endif', '//#endif')

                        if new_line.find(', *') != -1:
                            new_line = new_line.replace(', *', ';//, *')
                            enum_scope = False
                    else:
                        # INTERFACE macros only appear within the !CINTERFACE
                        # scope.
                        new_line = new_line.replace(
                            'MIDL_INTERFACE', 'class //MIDL_INTERFACE'
                        )
                        new_line = new_line.replace(
                            'BEGIN_INTERFACE', '//BEGIN_INTERFACE'
                        )
                        new_line = new_line.replace(
                            'END_INTERFACE', '//END_INTERFACE'
                        )
                    source += new_line
        return source



#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write


# Generates declarations for functions to unwrap struct members with a
# COM object type before the struct is passed as an argument to a DX12
# API call during DX12 capture.
class Dx12StructUnwrappersHeaderGenerator(Dx12BaseGenerator):
    # Default C++ code indentation size.
    INDENT_SIZE = 4

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )

    # Method override
    def beginFile(self, genOpts):
        Dx12BaseGenerator.beginFile(self, genOpts)

        self.write_include()

        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(encode)', file=self.outFile)

    # Method override
    def endFile(self):
        self.newline()

        # List containing names of structs with COM object members.
        structs_with_objects = {
            **self.CUSTOM_STRUCT_HANDLE_MAP, 'D3D12_CPU_DESCRIPTOR_HANDLE':
            ['ptr']
        }

        # Find structs with COM object members, which will need to be
        # unwrapped.
        struct_dict = self.source_dict['struct_dict']
        for struct in struct_dict:
            self.check_struct_member_handles(
                struct, structs_with_objects, None, True
            )

        # Generate unwrap functions for any structs that were added to
        # the list.
        for key in structs_with_objects:
            if not self.is_struct_black_listed(key):
                self.write_struct_unwrap_def(key)

        write('template <typename T>', file=self.outFile)
        write(
            'T* MakeUnwrapStructs(const T* values, size_t len, HandleUnwrapMemory* unwrap_memory)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write(
            '    assert((values != nullptr) && (len > 0) && (unwrap_memory != nullptr));',
            file=self.outFile
        )
        self.newline()
        write(
            '    const uint8_t* bytes     = reinterpret_cast<const uint8_t*>(values);',
            file=self.outFile
        )
        write(
            '    size_t         num_bytes = len * sizeof(T);',
            file=self.outFile
        )
        self.newline()
        write(
            '    return reinterpret_cast<T*>(unwrap_memory->GetFilledBuffer(bytes, num_bytes));',
            file=self.outFile
        )
        write('}', file=self.outFile)
        self.newline()
        write('template <typename T>', file=self.outFile)
        write(
            'const T* UnwrapStructPtrObjects(const T* value, HandleUnwrapMemory* unwrap_memory)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    T* unwrapped_struct = nullptr;', file=self.outFile)
        self.newline()
        write('    if (value != nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        unwrapped_struct = MakeUnwrapStructs(value, 1, unwrap_memory);',
            file=self.outFile
        )
        write(
            '        UnwrapStructObjects(unwrapped_struct, unwrap_memory);',
            file=self.outFile
        )
        write('    }', file=self.outFile)
        self.newline()
        write('    return unwrapped_struct;', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()
        write('template <typename T>', file=self.outFile)
        write(
            'const T* UnwrapStructArrayObjects(const T* values, size_t len, HandleUnwrapMemory* unwrap_memory)',
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    if ((values != nullptr) && (len > 0))', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        auto unwrapped_structs = MakeUnwrapStructs(values, len, unwrap_memory);',
            file=self.outFile
        )
        self.newline()
        write('        for (size_t i = 0; i < len; ++i)', file=self.outFile)
        write('        {', file=self.outFile)
        write(
            '            UnwrapStructObjects(&unwrapped_structs[i], unwrap_memory);',
            file=self.outFile
        )
        write('        }', file=self.outFile)
        self.newline()
        write('        return unwrapped_structs;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write(
            '    // Leave the original memory in place when the pointer is not null, but size is zero.',
            file=self.outFile
        )
        write('    return values;', file=self.outFile)
        write('}', file=self.outFile)
        self.newline()

        write('GFXRECON_END_NAMESPACE(encode)', file=self.outFile)
        write('GFXRECON_END_NAMESPACE(gfxrecon)', file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)

    # Method override
    def generate_feature(self):
        Dx12BaseGenerator.generate_feature(self)

    def write_struct_unwrap_def(self, struct, indent=''):
        expr = indent + 'void UnwrapStructObjects({}* value, '\
            'HandleUnwrapMemory* unwrap_memory);\n'.format(struct)

        write(expr, file=self.outFile)

    def write_include(self):
        code = ''

        code += '#include "encode/handle_unwrap_memory.h"\n'
        code += '#include "util/defines.h"\n'
        code += '\n'

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            code += '#include <{}>\n'.format(k)

        write(code, file=self.outFile)

    def increment_indent(self, indent):
        return indent + (' ' * self.INDENT_SIZE)

    def decrement_indent(self, indent):
        return indent[:-self.INDENT_SIZE]



#!/usr/bin/python3 -i

import sys
from vulkan_base_generator import VulkanBaseGenerator, VulkanBaseGeneratorOptions, write


class VulkanConstantMapsGeneratorOptions(VulkanBaseGeneratorOptions):
    """Adds the following new option:
    is_override - Specify whether the member function declarations are
                  virtual function overrides or pure virtual functions.
    Options for generating C++ class declarations for Vulkan parameter processing.
    """

    def __init__(
        self,
        blacklists=None,  # Path to JSON file listing apicalls and structs to ignore.
        platform_types=None,  # Path to JSON file listing platform (WIN32, X11, etc.) defined types.
        filename=None,
        directory='.',
        prefix_text='',
        protect_file=False,
        protect_feature=True,
        extra_headers=[]
    ):
        VulkanBaseGeneratorOptions.__init__(
            self,
            blacklists,
            platform_types,
            filename,
            directory,
            prefix_text,
            protect_file,
            protect_feature,
            extra_headers=extra_headers
        )

        self.begin_end_file_data.specific_headers.extend((
            'util/defines.h',
        ))
        self.begin_end_file_data.system_headers.extend((
            'unordered_map',
            'vector',
        ))
        self.begin_end_file_data.namespaces.extend(('gfxrecon', 'decode'))


class VulkanConstantMapsGenerator(VulkanBaseGenerator):
    """VulkanColorspaceMapGenerator - subclass of VulkanBaseGenerator.
    """

    def __init__(
        self, err_file=sys.stderr, warn_file=sys.stderr, diag_file=sys.stdout
    ):
        VulkanBaseGenerator.__init__(
            self,
            err_file=err_file,
            warn_file=warn_file,
            diag_file=diag_file
        )

    def beginFile(self, gen_opts):
        """Method override."""
        VulkanBaseGenerator.beginFile(self, gen_opts)

        self.newline()

        colorspace_extensions_names = set()
        colorspace_extension_map = dict.fromkeys(
            filter(
                lambda enum_value: 'VK_COLOR_SPACE' in enum_value,
                self.registry.enumdict
            )
        )
        colorspace_extension_map.pop(
            'VK_COLOR_SPACE_SRGB_NONLINEAR_KHR'
        )  # Remove the core provided color space from extension map
        for extension in self.registry.extensions:
            colorspace_enum = filter(
                lambda element: 'VK_COLOR_SPACE' in element.get('name'),
                extension.findall('require/enum')
            )
            colorspace_enum = map(
                lambda element: element.get('name'), colorspace_enum
            )
            for name in colorspace_enum:
                value = extension.get("name")
                colorspace_extensions_names.add(value)
                colorspace_extension_map[name] = f'"{value}"'

        write(
            'static const std::unordered_map<VkColorSpaceKHR, const char *> kColorSpaceExtensionMap = {',
            file=self.outFile
        )
        for key, value in colorspace_extension_map.items():
            write(f'{{{key}, {value}}},', file=self.outFile)
        write('};', file=self.outFile)
        self.newline()

        write(
            'static const std::vector<const char *> kColorSpaceExtensionNames = {',
            file=self.outFile
        )
        for extension_name in sorted(colorspace_extensions_names):
            write(f'"{extension_name}",', file=self.outFile)
        write('};', file=self.outFile)
        self.newline()



#!/usr/bin/python3
#
# Copyright 2018-2024, The Khronos Group Inc.
# Copyright (c) 2018-2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>
#
# Purpose:      This file contains tests for check_spec_links.py
#               that depend on the API being used.

import pytest
from check_spec_links import MessageId
from test_check_spec_links import CheckerWrapper, allMessages


@pytest.fixture
def ckr(capsys):
    """Fixture - add an arg named ckr to your test function to automatically get one passed to you."""
    return CheckerWrapper(capsys)


def test_openxr_refpage_mismatch(ckr):
    """Test the REFPAGE_MISMATCH message."""
    ckr.enabled([MessageId.REFPAGE_MISMATCH])
    # Should error: we undid the special case of the include for FlagBits inside of refpage for Flags
    assert ckr.check(
        """[open,refpage='XrSwapchainCreateFlags']
        --
        include::{generated}/api/enums/XrSwapchainCreateFlagBits.adoc[]""").messages

    # Should not error, this is how we expect it
    assert not ckr.check(
        """[open,refpage='XrSwapchainCreateFlagBits']
        --
        include::{generated}/api/enums/XrSwapchainCreateFlagBits.adoc[]""").messages

def test_openxr_refpage_type(ckr):
    """Test the REFPAGE_TYPE message, for fallout following the flags fix"""
    ckr.enabled([MessageId.REFPAGE_TYPE])
    # Should not error, this is how we expect it
    assert ckr.check(
        """[open,type='enums',refpage='XrEnvironmentBlendMode']
        --
        include::{generated}/api/enums/XrEnvironmentBlendMode.adoc[]""").numDiagnostics() == 0


def test_openxr_refpage_missing(ckr):
    """OpenXR-specific tests of the REFPAGE_MISSING message."""
    ckr.enabled([MessageId.REFPAGE_MISSING])

    # Should error: all flags includes are in the right place now so we need ref pages for them
    assert ckr.check(
        "include::{generated}/api/flags/XrSwapchainCreateFlags.adoc[]").messages
    assert ckr.check(
        "include::{generated}/api/flags/XrSwapchainCreateFlagBits.adoc[]").messages


def test_openxr_refpage_block(ckr):
    """OpenXR-specific tests of the REFPAGE_BLOCK message."""
    ckr.enabled([MessageId.REFPAGE_BLOCK])

    # Should have 1 error: line before '--' isn't tag
    assert ckr.check(
        """--
        bla
        --""").numDiagnostics() == 1

    # Should have 3 errors:
    #  - line before '--' isn't tag (refpage gets opened anyway),
    #  - tag is inside refpage (which auto-closes the previous refpage block, then re-opens a refpage)
    #  - line after tag isn't '--'
    result = ckr.check(
        """--
        [open,]
        bla
        --""")
    assert result.numDiagnostics() == 3
    # Internally, it's as if the following were the spec source, after putting in the "fake" lines
    # (each of the added lines comes from one message):
    #
    # [open,]
    # --
    # --
    # [open,]
    # --
    # bla
    # --
    assert len([x for x in allMessages(result)
                if "but did not find, a line containing only -- following a reference page tag" in x]) == 1
    assert len([x for x in allMessages(result)
                if "containing only -- outside of a reference page block" in x]) == 1
    assert len([x for x in allMessages(result)
                if "we are already in a refpage block" in x]) == 1



#!/usr/bin/python3 -i
#
# Copyright (c) 2019 Valve Corporation
# Copyright (c) 2019-2025 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import re, sys
from khronos_base_generator import write


class KhronosStructHandleWrappersBodyGenerator():
    """KhronosStructHandleWrappersBodyGenerator
    Generates C++ functions responsible for wrapping struct member handles
    when recording Khrosos API call parameter data.
    """

    def write_special_case_struct_handling(self):
        """Method may be overridden"""
        return

    def generate_parent_child_handling(self, api_data, type):
        body = ''
        has_case = False
        if type in self.children_structs:
            type_var_name = api_data.struct_type_variable
            body += '        switch (value->{})\n'.format(type_var_name)
            body += '        {\n'
            body += '            default:\n'
            body += '                // Handle as parent-type below\n'
            body += '                break;\n'

            # Loop over each possible child
            has_case = False
            for child in self.children_structs[type]:
                if child not in self.structs_with_handles:
                    continue
                has_case = True
                switch_type = self.struct_type_names[child]

                body += f'            case {switch_type}:\n'
                body += f'                UnwrapStructHandles(reinterpret_cast<{child}*>(value),\n'
                body += f'                                 unwrap_memory);\n'
                body += '                // Return here because we processed the appropriate data in\n'
                body += '                // the correct structure type\n'
                body += '                return;\n'
            body += '        }\n'
            body += '\n'

        return body if has_case else ''

    def write_struct_handle_wrapper_content(self):
        api_data = self.get_api_data()
        ext_struct_name = api_data.extended_struct_func_prefix
        ext_var_name = api_data.extended_struct_variable
        type_var_name = api_data.struct_type_variable
        const_prefix = ''
        if api_data.return_const_ptr_on_extended:
            const_prefix = 'const '

        for struct in self.get_all_filtered_struct_names():
            if (
                (struct in self.structs_with_handles) or
                self.child_struct_has_handles(struct) or
                (struct in self.GENERIC_HANDLE_STRUCTS)
            ) and (struct not in self.STRUCT_MAPPERS_BLACKLIST):
                handle_members = dict()
                generic_handle_members = dict()

                if struct in self.structs_with_handles:
                    handle_members = self.structs_with_handles[struct]
                if struct in self.GENERIC_HANDLE_STRUCTS:
                    generic_handle_members = self.GENERIC_HANDLE_STRUCTS[struct
                                                                         ]

                body = '\n'
                body += 'void UnwrapStructHandles({}* value, HandleUnwrapMemory* unwrap_memory)\n'.format(
                    struct
                )
                
                body += '{\n'

                unwrapping = self.generate_parent_child_handling(api_data, struct)
                unwrapping += self.make_struct_handle_unwrappings(api_data, struct, handle_members, generic_handle_members)
                
                if unwrapping:
                    body += '    if (value != nullptr)\n'
                    body += '    {\n'
                    body += unwrapping
                    body += '    }\n'

                body += '}'

                write(body, file=self.outFile)

        # Generate the pNext shallow copy code, for pNext structs that don't have handles, but need to be preserved in the overall copy for handle wrapping.
        self.newline()
        write(
            '{0}* Copy{1}Struct(const {0}* base, HandleUnwrapMemory* unwrap_memory)'
            .format(api_data.base_in_struct, ext_struct_name),
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    assert(base != nullptr);', file=self.outFile)
        self.newline()
        write(
            '    {}* copy = nullptr;'.format(api_data.base_in_struct),
            file=self.outFile
        )
        write('    switch (base->{})'.format(type_var_name), file=self.outFile)
        write('    {', file=self.outFile)
        write('    default:', file=self.outFile)
        write(
            '        GFXRECON_LOG_WARNING("Failed to copy entire {0} chain when unwrapping handles due to unrecognized {1} %d", base->{1});'
            .format(ext_var_name, type_var_name),
            file=self.outFile
        )
        write('        break;', file=self.outFile)
        self.write_special_case_struct_handling()

        extended_list = []
        for struct in self.all_extended_structs:
            for ext_struct in self.all_extended_structs[struct]:
                if ext_struct not in extended_list and ext_struct not in self.all_struct_aliases:
                    extended_list.append(ext_struct)

        for base_type in sorted(extended_list):
            if base_type not in self.struct_type_names:
                continue

            stype = self.struct_type_names[base_type]
            write('    case {}:'.format(stype), file=self.outFile)
            write(
                '        copy = reinterpret_cast<{}*>(MakeUnwrapStructs(reinterpret_cast<const {}*>(base), 1, unwrap_memory));'
                .format(api_data.base_in_struct, base_type),
                file=self.outFile
            )
            write('        break;', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write('    return copy;', file=self.outFile)
        write('}', file=self.outFile)

        # Generate the extended struct handle wrapping code.
        self.newline()
        write(
            '{const}void* Unwrap{struct}StructHandles(const void* value, HandleUnwrapMemory* unwrap_memory)'
            .format(const=const_prefix, struct=ext_struct_name),
            file=self.outFile
        )
        write('{', file=self.outFile)
        write('    if (value != nullptr)', file=self.outFile)
        write('    {', file=self.outFile)
        write(
            '        const {0}* base = reinterpret_cast<const {0}*>(value);'.
            format(api_data.base_in_struct),
            file=self.outFile
        )
        self.newline()
        write(
            '        switch (base->{})'.format(type_var_name),
            file=self.outFile
        )
        write('        {', file=self.outFile)
        write('        default:', file=self.outFile)
        write('        {', file=self.outFile)
        write(
            '            // This structure does not contain handles, but may point to a structure that does.',
            file=self.outFile
        )
        write(
            '            {}* copy = Copy{}Struct(base, unwrap_memory);'.format(
                api_data.base_in_struct, ext_struct_name
            ),
            file=self.outFile
        )
        write('            if (copy != nullptr)', file=self.outFile)
        write('            {', file=self.outFile)
        write(
            '                copy->{2} = reinterpret_cast<const {0}*>(Unwrap{1}StructHandles(base->{2}, unwrap_memory));'
            .format(api_data.base_in_struct, ext_struct_name, ext_var_name),
            file=self.outFile
        )
        write('            }', file=self.outFile)
        write('            return copy;', file=self.outFile)
        write('        }', file=self.outFile)
        for base_type in sorted(extended_list):
            if (
                base_type in self.structs_with_handles
                and base_type in self.struct_type_names
            ):
                stype = self.struct_type_names[base_type]
                write('        case {}:'.format(stype), file=self.outFile)
                write(
                    '            return UnwrapStructPtrHandles(reinterpret_cast<const {}*>(base), unwrap_memory);'
                    .format(base_type),
                    file=self.outFile
                )
        write('        }', file=self.outFile)
        write('    }', file=self.outFile)
        self.newline()
        write('    return nullptr;', file=self.outFile)
        write('}', file=self.outFile)

    def has_special_case_handle_unwrapping(self, name):
        """Method may be overridden."""
        return False

    def get_special_case_handle_wrapping(self, name):
        """Method may be overridden."""
        return

    def make_struct_handle_unwrappings(
        self, api_data, name, handle_members, generic_handle_members
    ):
        """Generating expressions for unwrapping struct handles before an API call."""
        body = ''

        for member in handle_members:
            if member.name == api_data.extended_struct_variable:
                body += '        if (value->{} != nullptr)\n'.format(
                    member.name
                )
                body += '        {\n'
                if self.has_special_case_handle_unwrapping(name):
                    body += self.get_special_case_handle_wrapping(name)
                else:
                    body += '            value->{0} = Unwrap{1}StructHandles(value->{0}, unwrap_memory);\n'.format(
                        member.name, api_data.extended_struct_func_prefix
                    )
                body += '        }\n'
            elif self.is_struct(member.base_type):
                # This is a struct that includes handles.
                if member.is_array:
                    if api_data.return_const_ptr_on_extended:
                        body += '        value->{name} = UnwrapStructArrayHandles(value->{name}, value->{}, unwrap_memory);\n'.format(
                            member.array_length, name=member.name
                        )
                    else:
                        if 'const' in member.full_type:
                            temp_type = member.full_type
                            full_type_non_const = re.sub(' const ', '', temp_type)
                            full_type_non_const = re.sub(
                                'const ', '', full_type_non_const
                            )
                            full_type_non_const = re.sub(
                                ' const', '', full_type_non_const
                            )
                            variable_name = f'const_cast<{full_type_non_const}>(value->{member.name})'
                        else:
                            variable_name = f'value->{member.name}'

                        length_exprs = member.array_length.split(',')
                        length_count = len(length_exprs)

                        if member.pointer_count > 1 and length_count < member.pointer_count:
                            unwrap_function = 'UnwrapStructPtrArrayHandles'
                        else:
                            unwrap_function = 'UnwrapStructArrayHandles'

                        body += f'        value->{member.name} = {unwrap_function}({variable_name}, value->{member.array_length}, unwrap_memory);\n'
                elif member.is_pointer:
                    body += '        value->{name} = UnwrapStructPtrHandles(value->{name}, unwrap_memory);\n'.format(
                        name=member.name
                    )
                else:
                    body += '        UnwrapStructHandles(&value->{}, unwrap_memory);\n'.format(
                        member.name
                    )
        return body



#!/usr/bin/env python3
#
# Copyright (c) 2021 LunarG, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to
# deal in the Software without restriction, including without limitation the
# rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
# sell copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
# FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
# IN THE SOFTWARE.

import sys
from dx12_base_generator import Dx12BaseGenerator, write
from dx12_base_struct_decoders_header_generator import Dx12BaseStructDecodersHeaderGenerator


class Dx12StructDecodersHeaderGenerator(
    Dx12BaseGenerator, Dx12BaseStructDecodersHeaderGenerator
):
    """Generates C++ functions responsible for decoding Dx12 API calls."""

    def __init__(
        self,
        source_dict,
        dx12_prefix_strings,
        err_file=sys.stderr,
        warn_file=sys.stderr,
        diag_file=sys.stdout
    ):
        Dx12BaseGenerator.__init__(
            self, source_dict, dx12_prefix_strings, err_file, warn_file,
            diag_file
        )
        self.check_blacklist = True

    def beginFile(self, gen_opts):
        """Method override."""
        Dx12BaseGenerator.beginFile(self, gen_opts)
        write('#if defined(D3D12_SUPPORT)', file=self.outFile)
        self.newline()

        self.write_include()
        write('GFXRECON_BEGIN_NAMESPACE(gfxrecon)', file=self.outFile)
        write('GFXRECON_BEGIN_NAMESPACE(decode)', file=self.outFile)
        self.newline()

    def generate_feature(self):
        """Method override."""
        Dx12BaseGenerator.generate_feature(self)
        Dx12BaseStructDecodersHeaderGenerator.generate_feature(self)

    def write_include(self):
        """Method override."""
        code = ''
        code += '#ifdef WIN32\n'

        header_dict = self.source_dict['header_dict']
        for k, v in header_dict.items():
            code += '#include <{}>\n'.format(k)

        code += '#else\n'
        code += '#include "format/platform_types.h"\n'
        code += '#endif // WIN32\n\n'

        code += (
            "#include \"decode/custom_dx12_struct_decoders_forward.h\"\n"
            "#include \"decode/handle_pointer_decoder.h\"\n"
            "#include \"decode/string_array_decoder.h\"\n"
            "#include \"decode/string_decoder.h\"\n"
            "#include \"decode/struct_pointer_decoder.h\"\n"
            "#include \"format/format.h\"\n"
            "#include \"generated/generated_dx12_struct_decoders_forward.h\"\n"
            "#include \"util/defines.h\"\n"
        )
        write(code, file=self.outFile)

    def endFile(self):
        """Method override."""
        code = '\n'
        code += 'GFXRECON_END_NAMESPACE(decode)\n'
        code += 'GFXRECON_END_NAMESPACE(gfxrecon)\n'
        code += '\n'
        code += '#endif // defined(D3D12_SUPPORT)'
        write(code, file=self.outFile)

        # Finish processing in superclass
        Dx12BaseGenerator.endFile(self)



#!/usr/bin/python3 -i
#
# Copyright (c) 2019-2024, The Khronos Group Inc.
# Copyright (c) 2019 Collabora, Ltd.
#
# SPDX-License-Identifier: Apache-2.0
#
# Author(s):    Rylie Pavlik <rylie.pavlik@collabora.com>
"""Provides utilities to write a script to verify XML registry consistency."""

import re
from typing import Set

import networkx as nx
from networkx.algorithms import shortest_path

from .algo import RecursiveMemoize
from .attributes import ExternSyncEntry, LengthEntry
from .data_structures import DictOfStringSets
from .util import findNamedElem, getElemName, getElemType
from .conventions import ConventionsBase


def _get_extension_tags(reg):
    """Get a set of all author tags registered for use."""
    return set(elt.get("name") for elt in reg.tree.findall("./tags/tag[@name]"))


class XMLChecker:
    def __init__(self, entity_db,  conventions: ConventionsBase, manual_types_to_codes=None,
                 forward_only_types_to_codes=None,
                 reverse_only_types_to_codes=None,
                 suppressions=None,
                 display_warnings=True):
        """Set up data structures.

        May extend - call:
        `super().__init__(db, conventions, manual_types_to_codes)`
        as the last statement in your function.

        manual_types_to_codes is a dictionary of hard-coded
        "manual" return codes:
        the codes of the value are available for a command if-and-only-if
        the key type is passed as an input.

        forward_only_types_to_codes is additional entries to the above
        that should only be used in the "forward" direction
        (arg type implies return code)

        reverse_only_types_to_codes is additional entries to
        manual_types_to_codes that should only be used in the
        "reverse" direction
        (return code implies arg type)
        """
        self.fail = False
        self.entity = None
        self.errors = DictOfStringSets()
        self.warnings = DictOfStringSets()
        self.db = entity_db
        self.reg = entity_db.registry
        self.handle_data = HandleData(self.reg)
        self.conventions = conventions
        self.display_warnings = display_warnings

        self.CONST_RE = re.compile(r"\bconst\b")
        self.ARRAY_RE = re.compile(r"\[[^]]+\]")

        # Init memoized properties
        self._handle_data = None

        if not manual_types_to_codes:
            manual_types_to_codes = {}
        if not reverse_only_types_to_codes:
            reverse_only_types_to_codes = {}
        if not forward_only_types_to_codes:
            forward_only_types_to_codes = {}

        reverse_codes = DictOfStringSets(reverse_only_types_to_codes)
        forward_codes = DictOfStringSets(forward_only_types_to_codes)
        for k, v in manual_types_to_codes.items():
            forward_codes.add(k, v)
            reverse_codes.add(k, v)

        self.forward_only_manual_types_to_codes = forward_codes.get_dict()
        self.reverse_only_manual_types_to_codes = reverse_codes.get_dict()

        # The presence of some types as input to a function imply the
        # availability of some return codes.
        self.input_type_to_codes = compute_type_to_codes(
            self.handle_data,
            forward_codes,
            extra_op=self.add_extra_codes)

        # Some return codes require a type (or its child) in the input.
        self.codes_requiring_input_type = compute_codes_requiring_type(
            self.handle_data,
            reverse_codes
        )

        specified_codes = set(self.codes_requiring_input_type.keys())
        for codes in self.forward_only_manual_types_to_codes.values():
            specified_codes.update(codes)
        for codes in self.reverse_only_manual_types_to_codes.values():
            specified_codes.update(codes)
        for codes in self.input_type_to_codes.values():
            specified_codes.update(codes)

        self.return_codes: Set[str]
        unrecognized = specified_codes - self.return_codes
        if unrecognized:
            raise RuntimeError("Return code mentioned in script that isn't in the registry: " +
                               ', '.join(unrecognized))

        self.referenced_input_types = ReferencedTypes(self.db, self.is_input)
        self.referenced_types = ReferencedTypes(self.db)
        if not suppressions:
            suppressions = {}
        self.suppressions = DictOfStringSets(suppressions)
        self.tags = _get_extension_tags(self.db.registry)

    def is_api_type(self, member_elem):
        """Return true if the member/parameter ElementTree passed is from this API.

        May override or extend."""
        membertext = "".join(member_elem.itertext())

        return self.conventions.type_prefix in membertext

    def is_input(self, member_elem):
        """Return true if the member/parameter ElementTree passed is
        considered "input".

        May override or extend."""
        membertext = "".join(member_elem.itertext())

        if self.conventions.type_prefix not in membertext:
            return False

        ret = True
        # Const is always input.
        if self.CONST_RE.search(membertext):
            ret = True

        # Arrays and pointers that aren't const are always output.
        elif "*" in membertext:
            ret = False
        elif self.ARRAY_RE.search(membertext):
            ret = False

        return ret

    def strip_extension_tag(self, name):
        """Remove a single author tag from the end of a name, if any.

        Returns the stripped name and the tag, or the input and None if there was no tag.
        """
        # Author tag can be suffixed with experimental version
        name_no_experimental = re.sub("X[0-9]*$", "", name)

        for t in self.tags:
            if (
                self.conventions.allows_x_number_suffix
                and name_no_experimental.endswith(t)
            ):
                name = name_no_experimental

            if name.endswith(t):
                name = name[:-(len(t))]
                if name[-1] == "_":
                    # remove trailing underscore
                    name = name[:-1]
                return name, t
        return name, None

    def add_extra_codes(self, types_to_codes):
        """Add any desired entries to the types-to-codes DictOfStringSets
        before performing "ancestor propagation".

        Passed to compute_type_to_codes as the extra_op.

        May override."""
        pass

    def should_skip_checking_codes(self, name):
        """Return True if more than the basic validation of return codes should
        be skipped for a command.

        May override."""

        return self.conventions.should_skip_checking_codes

    def get_codes_for_command_and_type(self, cmd_name, type_name):
        """Return a set of return codes expected due to having
        an input argument of type type_name.

        The cmd_name is passed for use by extending methods.
        Note that you should not use cmd_name to add codes, just to
        filter them out. See get_required_codes_for_command() to do that.

        May extend."""
        return self.input_type_to_codes.get(type_name, set())

    def get_required_codes_for_command(self, cmd_name):
        """Return a set of return codes required due to having a particular name.

        May override."""
        return set()

    def get_forbidden_codes_for_command(self, cmd_name):
        """Return a set of return codes not permittted due to having a particular name.

        May override."""
        return set()

    def check(self):
        """Iterate through the registry, looking for consistency problems.

        Outputs error messages at the end."""
        # Iterate through commands, looking for consistency problems.
        for name, info in self.reg.cmddict.items():
            self.set_error_context(entity=name, elem=info.elem)

            self.check_command(name, info)

        for name, info in self.reg.typedict.items():
            cat = info.elem.get('category')
            if not cat:
                # This is an external thing, skip it.
                continue
            self.set_error_context(entity=name, elem=info.elem)

            self.check_type(name, info, cat)

        self.ext_numbers = set()
        for name, info in self.reg.extdict.items():
            self.set_error_context(entity=name, elem=info.elem)

            # Determine if this extension is supported by the API we're
            # testing, and pass that flag to check_extension.
            # For Vulkan, multiple APIs can be specified in the 'supported'
            # attribute.
            supported_apis = info.elem.get('supported', '').split(',')
            supported = self.conventions.xml_api_name in supported_apis
            self.check_extension(name, info, supported)

        self.check_format()

        entities_with_messages = set(
            self.errors.keys()).union(self.warnings.keys())
        if entities_with_messages:
            print('xml_consistency/consistency_tools error and warning messages follow.')

        for entity in entities_with_messages:
            print()
            print('-------------------')
            print('Messages for', entity)
            print()
            messages = self.errors.get(entity)
            if messages:
                for m in messages:
                    print('Error:', m)

            messages = self.warnings.get(entity)
            if messages and self.display_warnings:
                for m in messages:
                    print('Warning:', m)

    def check_param(self, param):
        """Check a member of a struct or a param of a function.

        Called from check_params.

        May extend."""
        param_name = getElemName(param)
        # Make sure there's something between the type and the name
        # Can't just look at the .tail of <type> for some reason,
        # so instead we look to see if anything's between
        # type's text and name's text in the itertext.
        # If there's no text between the tags, there will be no string
        # between those tags' text in itertext()
        text_parts = list(param.itertext())
        type_idx = text_parts.index(getElemType(param))
        name_idx = text_parts.index(param_name)
        if name_idx - type_idx == 1:
            self.record_error(
                "Space (or other delimiter text) missing between </type> and <name> for param/member named",
                param_name)

        # Check external sync entries
        externsyncs = ExternSyncEntry.parse_externsync_from_param(param)
        if externsyncs:
            for entry in externsyncs:
                if entry.entirely_extern_sync:
                    if len(externsyncs) > 1:
                        self.record_error("Comma-separated list in externsync attribute includes 'true' for",
                                          param_name)
                else:
                    # member name
                    # TODO only looking at the superficial feature here,
                    # not entry.param_ref_parts
                    if entry.member != param_name:
                        self.record_error("externsync attribute for", param_name,
                                          "refers to some other member/parameter:", entry.member)

    def check_params(self, params):
        """Check the members of a struct or params of a function.

        Called from check_type and check_command.

        May extend."""
        for param in params:
            self.check_param(param)

            # Check for parameters referenced by len= attribute
            lengths = LengthEntry.parse_len_from_param(param)
            if lengths:
                for entry in lengths:
                    if not entry.other_param_name:
                        continue
                    # TODO only looking at the superficial feature here,
                    # not entry.param_ref_parts
                    other_param = findNamedElem(params, entry.other_param_name)
                    if other_param is None:
                        self.record_error("References a non-existent parameter/member in the length of",
                                          getElemName(param), ":", entry.other_param_name)

    def check_referenced_type(self, desc, ref_name):
        """
        Record an error if a type mentioned somewhere doesn't exist.

        :param desc: Description of where this type reference was found,
                     for the error message.
        :param ref_name: The name of the referenced type. If false-ish (incl. None),
                         checking is skipped, so OK to pass the results of
                         info.elem.get() directly
        """
        if ref_name:
            entity = self.db.findEntity(ref_name)
            if not entity:
                self.record_error("Unknown type named in", desc, ":",
                                  ref_name)

    def check_type(self, name, info, category):
        """Check a type's XML data for consistency.

        Called from check.

        May extend."""
        if category == 'struct':
            if not name.startswith(self.conventions.type_prefix):
                self.record_error("Name does not start with",
                                  self.conventions.type_prefix)
            members = info.elem.findall('member')
            self.check_params(members)

            # Check the structure type member, if present.
            type_member = findNamedElem(
                members, self.conventions.structtype_member_name)
            if type_member is not None:
                val = type_member.get('values')
                if val:
                    expected = self.conventions.generate_structure_type_from_name(
                        name)
                    if val != expected:
                        self.record_error("Type has incorrect type-member value: expected",
                                          expected, "got", val)

            # Check structextends attribute, if present.
            # For Vulkan, this may be a comma-separated list of multiple types
            for type in info.elem.get("structextends", '').split(','):
                self.check_referenced_type("'structextends' attribute", type)

            # Check parentstruct attribute, if present.
            self.check_referenced_type("'parentstruct' attribute", info.elem.get("parentstruct"))

        elif category == "bitmask":
            if 'Flags' not in name:
                self.record_error("Name of bitmask doesn't include 'Flags'")
        elif category == "handle":
            # Check parent attribute, if present.
            self.check_referenced_type("'parent' attribute", info.elem.get("parent"))

    def check_extension(self, name, info, supported):
        """Check an extension's XML data for consistency.

        Called from check.

        May extend."""

        # Verify that each extension has a unique number
        extension_number = info.elem.get('number')
        if extension_number is not None and extension_number != '0':
            if extension_number in self.ext_numbers:
                self.record_error(f"Duplicate extension number {extension_number}")
            else:
                self.ext_numbers.add(extension_number)

    def check_format(self):
        """Check an extension's XML data for consistency.

        Called from check.

        May extend."""
        pass

    def check_command(self, name, info):
        """Check a command's XML data for consistency.

        Called from check.

        May extend."""
        elem = info.elem

        self.check_params(elem.findall('param'))

        # Some minimal return code checking
        errorcodes = elem.get("errorcodes")
        if errorcodes:
            errorcodes = errorcodes.split(",")
        else:
            errorcodes = []

        successcodes = elem.get("successcodes")
        if successcodes:
            successcodes = successcodes.split(",")
        else:
            successcodes = []

        if not successcodes and not errorcodes:
            # Early out if no return codes.
            return

        # Create a set for each group of codes, and check that
        # they aren't duplicated within or between groups.
        errorcodes_set = set(errorcodes)
        if len(errorcodes) != len(errorcodes_set):
            self.record_error("Contains a duplicate in errorcodes")

        successcodes_set = set(successcodes)
        if len(successcodes) != len(successcodes_set):
            self.record_error("Contains a duplicate in successcodes")

        if not successcodes_set.isdisjoint(errorcodes_set):
            self.record_error("Has errorcodes and successcodes that overlap")

        self.check_command_return_codes_basic(
            name, info, successcodes_set, errorcodes_set)

        # Continue to further return code checking if not "complicated"
        if not self.should_skip_checking_codes(name):
            codes_set = successcodes_set.union(errorcodes_set)
            self.check_command_return_codes(
                name, info, successcodes_set, errorcodes_set, codes_set)

    def check_command_return_codes_basic(self, name, info,
                                         successcodes, errorcodes):
        """Check a command's return codes for consistency.

        Called from check_command on every command.

        May extend."""

        # Check that all error codes include _ERROR_,
        #  and that no success codes do.
        for code in errorcodes:
            if "_ERROR_" not in code:
                self.record_error(
                    code, "in errorcodes but doesn't contain _ERROR_")

        for code in successcodes:
            if "_ERROR_" in code:
                self.record_error(code, "in successcodes but contain _ERROR_")

    def check_command_return_codes(self, name, type_info,
                                   successcodes, errorcodes,
                                   codes):
        """Check a command's return codes in-depth for consistency.

        Called from check_command, only if
        `self.should_skip_checking_codes(name)` is False.

        May extend."""
        referenced_input = self.referenced_input_types[name]
        referenced_types = self.referenced_types[name]
        error_prefix = f"{self.conventions.api_prefix}ERROR"

        bad_success = {x for x in successcodes if x.startswith(error_prefix)}
        if bad_success:
            self.record_error("Found error code(s)",
                              ",".join(bad_success),
                              "listed in the successcodes attributes")

        bad_errors = {x for x in errorcodes if not x.startswith(error_prefix)}
        if bad_errors:
            self.record_error("Found success code(s)",
                              ",".join(bad_errors),
                              "listed in the errorcodes attributes")

        # Check that we have all the codes we expect, based on input types.
        for referenced_type in referenced_input:
            required_codes = self.get_codes_for_command_and_type(
                name, referenced_type)
            missing_codes = required_codes - codes
            if missing_codes:
                path = self.referenced_input_types.shortest_path(
                    name, referenced_type)
                path_str = " -> ".join(path)
                self.record_error("Missing expected return code(s)",
                                  ",".join(missing_codes),
                                  "implied because of input of type",
                                  referenced_type,
                                  "found via path",
                                  path_str)

        # Check that we have all the codes we expect based on command name.
        missing_codes = self.get_required_codes_for_command(name) - codes
        if missing_codes:
            self.record_error("Missing expected return code(s)",
                              ",".join(missing_codes),
                              "implied because of the name of this command")

        # Check that we don't have any codes forbidden based on command name.
        forbidden = self.get_forbidden_codes_for_command(name).intersection(codes)
        if forbidden:
            self.record_error("Got return code(s)",
                              ", ".join(forbidden),
                              "that were forbidden due to the name of this command")

        # Check that, for each code returned by this command that we can
        # associate with a type, we have some type that can provide it.
        # e.g. can't have INSTANCE_LOST without an Instance
        # (or child of Instance).
        for code in codes:

            required_types = self.codes_requiring_input_type.get(code)
            if not required_types:
                # This code doesn't have a known requirement
                continue

            # TODO: do we look at referenced_types or referenced_input here?
            # the latter is stricter
            if not referenced_types.intersection(required_types):
                self.record_error("Unexpected return code", code,
                                  "- none of these types:",
                                  required_types,
                                  "found in the set of referenced types",
                                  referenced_types)

    ###
    # Utility properties/methods
    ###

    def set_error_context(self, entity=None, elem=None):
        """Set the entity and/or element for future record_error calls."""
        self.entity = entity
        self.elem = elem
        self.name = getElemName(elem)
        self.entity_suppressions = self.suppressions.get(getElemName(elem))

    def record_error(self, *args, **kwargs):
        """Record failure and an error message for the current context."""
        message = " ".join((str(x) for x in args))

        if self._is_message_suppressed(message):
            return

        message = self._prepend_sourceline_to_message(message, **kwargs)
        self.fail = True
        self.errors.add(self.entity, message)

    def record_warning(self, *args, **kwargs):
        """Record a warning message for the current context."""
        message = " ".join((str(x) for x in args))

        if self._is_message_suppressed(message):
            return

        message = self._prepend_sourceline_to_message(message, **kwargs)
        self.warnings.add(self.entity, message)

    def _is_message_suppressed(self, message):
        """Return True if the given message, for this entity, should be suppressed."""
        if not self.entity_suppressions:
            return False
        for suppress in self.entity_suppressions:
            if suppress in message:
                return True

        return False

    def _prepend_sourceline_to_message(self, message, **kwargs):
        """Prepend a file and/or line reference to the message, if possible.

        If filename is given as a keyword argument, it is used on its own.

        If filename is not given, this will attempt to retrieve the filename and line from an XML element.
        If 'elem' is given as a keyword argument and is not None, it is used to find the line.
        If 'elem' is given as None, no XML elements are looked at.
        If 'elem' is not supplied, the error context element is used.

        If using XML, the filename, if available, is retrieved from the Registry class.
        If using XML and python-lxml is installed, the source line is retrieved from whatever element is chosen."""
        fn = kwargs.get('filename')
        sourceline = None

        if fn is None:
            elem = kwargs.get('elem', self.elem)
            if elem is not None:
                sourceline = getattr(elem, 'sourceline', None)
                if self.reg.filename:
                    fn = self.reg.filename

        if fn is None and sourceline is None:
            return message

        if fn is None:
            return f"Line {sourceline}: {message}"

        if sourceline is None:
            return f"{fn}: {message}"

        return f"{fn}:{sourceline}: {message}"


class HandleParents(RecursiveMemoize):
    def __init__(self, handle_types):
        self.handle_types = handle_types

        def compute(handle_type):
            immediate_parent = self.handle_types[handle_type].elem.get(
                'parent')

            if immediate_parent is None:
                # No parents, no need to recurse
                return []

            # Support multiple (alternate) parents
            immediate_parents = immediate_parent.split(',')

            # Recurse, combine, and return
            all_parents = immediate_parents[:]
            for parent in immediate_parents:
                all_parents.extend(self[parent])
            return all_parents

        super().__init__(compute, handle_types.keys())


def _always_true(x):
    return True


class ReferencedTypes(RecursiveMemoize):
    """Find all types(optionally matching a predicate) that are referenced
    by a struct or function, recursively."""

    def __init__(self, db, predicate=None):
        """Initialize.

        Provide an EntityDB object and a predicate function."""
        self.db = db

        self.predicate = predicate
        if not self.predicate:
            # Default predicate is "anything goes"
            self.predicate = _always_true

        self._directly_referenced = {}
        self.graph = nx.DiGraph()

        def compute(type_name):
            """Compute and return all types referenced by type_name, recursively, that satisfy the predicate.

            Called by the [] operator in the base class."""
            types = self.directly_referenced(type_name)
            if not types:
                return types

            all_types = set()
            all_types.update(types)
            for t in types:
                referenced = self[t]
                if referenced is not None:
                    # If not leading to a cycle
                    all_types.update(referenced)
            return all_types

        # Initialize base class
        super().__init__(compute, permit_cycles=True)

    def shortest_path(self, source, target):
        """Get the shortest path between one type/function name and another."""
        # Trigger computation
        _ = self[source]

        return shortest_path(self.graph, source=source, target=target)

    def directly_referenced(self, type_name):
        """Get all types referenced directly by type_name that satisfy the predicate.

        Memoizes its results."""
        if type_name not in self._directly_referenced:
            members = self.db.getMemberElems(type_name)
            if members:
                types = ((member, member.find("type")) for member in members)
                self._directly_referenced[type_name] = set(type_elem.text for (member, type_elem) in types
                                                           if type_elem is not None and self.predicate(member))

            else:
                self._directly_referenced[type_name] = set()
            children = self.db.childTypes(type_name)
            if children:
                self._directly_referenced[type_name].update(children)
            # Update graph
            self.graph.add_node(type_name)
            self.graph.add_edges_from((type_name, t)
                                      for t in self._directly_referenced[type_name])

        return self._directly_referenced[type_name]


class HandleData:
    """Data about all the handle types available in an API specification."""

    def __init__(self, registry):
        self.reg = registry
        self._handle_types = None
        self._ancestors = None
        self._descendants = None

    @property
    def handle_types(self):
        """Return a dictionary of handle type names to type info."""
        if not self._handle_types:
            # First time requested - compute it.
            self._handle_types = {
                type_name: type_info
                for type_name, type_info in self.reg.typedict.items()
                if type_info.elem.get('category') == 'handle'
            }
        return self._handle_types

    @property
    def ancestors_dict(self):
        """Return a dictionary of handle type names to sets of ancestors."""
        if not self._ancestors:
            # First time requested - compute it.
            self._ancestors = HandleParents(self.handle_types).get_dict()
        return self._ancestors

    @property
    def descendants_dict(self):
        """Return a dictionary of handle type names to sets of descendants."""
        if not self._descendants:
            # First time requested - compute it.

            handle_parents = self.ancestors_dict

            def get_descendants(handle):
                return set(h for h in handle_parents.keys()
                           if handle in handle_parents[h])

            self._descendants = {
                h: get_descendants(h)
                for h in handle_parents.keys()
            }
        return self._descendants


def compute_type_to_codes(handle_data, types_to_codes, extra_op=None):
    """Compute a DictOfStringSets of input type to required return codes.

    - handle_data is a HandleData instance.
    - d is a dictionary of type names to strings or string collections of
      return codes.
    - extra_op, if any, is called after populating the output from the input
      dictionary, but before propagation of parent codes to child types.
      extra_op is called with the in-progress DictOfStringSets.

    Returns a DictOfStringSets of input type name to set of required return
    code names.
    """
    # Initialize with the supplied "manual" codes
    types_to_codes = DictOfStringSets(types_to_codes)

    # Dynamically generate more codes, if desired
    if extra_op:
        extra_op(types_to_codes)

    # Final post-processing

    # Any handle can result in its parent handle's codes too.

    handle_ancestors = handle_data.ancestors_dict

    extra_handle_codes = {}
    for handle_type, ancestors in handle_ancestors.items():
        # The sets of return codes corresponding to each ancestor type.
        ancestors_codes = [types_to_codes.get(ancestor, set())
                           for ancestor in ancestors]
        extra_handle_codes[handle_type] = set().union(*ancestors_codes)

    for handle_type, extras in extra_handle_codes.items():
        types_to_codes.add(handle_type, extras)

    return types_to_codes


def compute_codes_requiring_type(handle_data, types_to_codes, registry=None):
    """Compute a DictOfStringSets of return codes to a set of input types able
    to provide the ability to generate that code.

    handle_data is a HandleData instance.
    d is a dictionary of input types to associated return codes(same format
    as for input to compute_type_to_codes, may use same dict).
    This will invert that relationship, and also permit any "child handles"
    to satisfy a requirement for a parent in producing a code.

    Returns a DictOfStringSets of return code name to the set of parameter
    types that would allow that return code.
    """
    # Use DictOfStringSets to normalize the input into a dict with values
    # that are sets of strings
    in_dict = DictOfStringSets(types_to_codes)

    handle_descendants = handle_data.descendants_dict

    out = DictOfStringSets()
    for in_type, code_set in in_dict.items():
        descendants = handle_descendants.get(in_type)
        for code in code_set:
            out.add(code, in_type)
            if descendants:
                out.add(code, descendants)

    return out



# Copyright 2010 Baptiste Lepilleur and The JsonCpp Authors
# Distributed under MIT license, or public domain if desired and
# recognized in your jurisdiction.
# See file LICENSE for detail or copy at http://jsoncpp.sourceforge.net/LICENSE

from __future__ import print_function
import os.path
import sys

def fix_source_eol(path, is_dry_run = True, verbose = True, eol = '\n'):
    """Makes sure that all sources have the specified eol sequence (default: unix)."""
    if not os.path.isfile(path):
        raise ValueError('Path "%s" is not a file' % path)
    try:
        f = open(path, 'rb')
    except IOError as msg:
        print("%s: I/O Error: %s" % (file, str(msg)), file=sys.stderr)
        return False
    try:
        raw_lines = f.readlines()
    finally:
        f.close()
    fixed_lines = [line.rstrip('\r\n') + eol for line in raw_lines]
    if raw_lines != fixed_lines:
        print('%s =>' % path, end=' ')
        if not is_dry_run:
            f = open(path, "wb")
            try:
                f.writelines(fixed_lines)
            finally:
                f.close()
        if verbose:
            print(is_dry_run and ' NEED FIX' or ' FIXED')
    return True
##
##
##
##def _do_fix(is_dry_run = True):
##    from waftools import antglob
##    python_sources = antglob.glob('.',
##        includes = '**/*.py **/wscript **/wscript_build',
##        excludes = antglob.default_excludes + './waf.py',
##        prune_dirs = antglob.prune_dirs + 'waf-* ./build')
##    for path in python_sources:
##        _fix_python_source(path, is_dry_run)
##
##    cpp_sources = antglob.glob('.',
##        includes = '**/*.cpp **/*.h **/*.inl',
##        prune_dirs = antglob.prune_dirs + 'waf-* ./build')
##    for path in cpp_sources:
##        _fix_source_eol(path, is_dry_run)
##
##
##def dry_fix(context):
##    _do_fix(is_dry_run = True)
##
##def fix(context):
##    _do_fix(is_dry_run = False)
##
##def shutdown():
##    pass
##
##def check(context):
##    # Unit tests are run when "check" target is used
##    ut = UnitTest.unit_test()
##    ut.change_to_testfile_dir = True
##    ut.want_to_see_test_output = True
##    ut.want_to_see_test_error = True
##    ut.run()
##    ut.print_results()
